// ASTGEN: #include "a.h"

#include <vector>
using std::vector;

#define V(x) static_cast<void*>(x)
#undef nil
#define nil NULL

struct AstToken;

typedef struct Ast Ast;
typedef struct AttrBlock AttrBlock;
typedef struct Expr Expr;
typedef struct Stmt Stmt;
typedef struct Type Type;
typedef struct Structelem Structelem;
typedef struct Clause Clause;
typedef struct Pattern Pattern;
typedef struct PatternCompileState PatternCompileState;
typedef struct Yystype Yystype;
typedef struct Fnarg Fnarg;
typedef union Datum Datum;
typedef struct Thread Thread;
typedef void Code(Thread*);
typedef struct Codebuf Codebuf;
typedef struct Label Label;
typedef struct LabelStub LabelStub;
typedef struct Frame Frame;
typedef struct Scope Scope;
typedef struct Var Var;
typedef struct Fn Fn;
typedef struct Compile Compile;
typedef struct PatternXForm PatternXForm;
typedef struct Loc Loc;
typedef struct GTerm GTerm;
typedef struct PGram PGram;
typedef struct PToken PToken;
typedef struct ExtStruct ExtStruct;
typedef struct Array Array;
typedef struct AssocList AssocList;
typedef struct ListNode ListNode;
typedef struct Map Map;
typedef struct MapIter MapIter;
typedef struct ZLine ZLine;
typedef struct View View;
struct Typecheck;

extern Arena *asyntax;

// ASTGEN ARENA: asyntax

typedef enum CallOp
{
	CallRegular = 0,
	CallCtor,
} CallOp;
#include "expr.h"

enum
{
	FnExtend       = 1<<0,
	FnExtensible   = 1<<1,
	FnAttribute    = 1<<2,
	DeclareConst   = 1<<3,
};
#include "stmt.h"

struct Clause
{
	Pattern *pat;
	Stmt *body;
};
Clause *mkClause(Pattern *pat, Stmt *body);

typedef enum PatternOp
{
	PatternTag = 1,
	PatternDtor,
	PatternEmptyList,
	PatternList,
	PatternParse,
	PatternConst,
	PatternLit,
	PatternAny,
	PatternDisjunct,
} PatternOp;
struct Pattern
{
	Line line;
	PatternOp op;
	List<Name> *tags;
	Name name;
	int isptr;
	Pattern *subpat;
	Pattern *subpat2;
	List<Pattern*> *subpats;
	Expr *value;
	ExprParse *valuep;
	Var *var;
	Type *type;		// The type of value being matched
	Type *tagtype;		// Tag struct (sub-struct of `type')
				// for typecasting in tag match
};
Pattern *mkPatternTag(List<Name> *tags, Pattern *subpat);
Pattern *mkPatternDtor(List<Pattern*> *subpats);
Pattern *mkPatternEmptyList();
Pattern *mkPatternList(Pattern *subpat, Pattern *subpat2);
Pattern *mkPatternParse(ExprParse *valuep, Name name);
Pattern *mkPatternConst(Expr *value);
Pattern *mkPatternLit(Name name, int isptr);
Pattern *mkPatternAny();
Pattern *mkPatternDisjunct(Pattern *subpat, Pattern *subpat2);

struct PatternCompileState
{
	// Note that newly bound variables must be _prepended_ to
	// boundVars so a previous boundVars list can be restored by
	// simply restoring the value of the old list pointer.
	List<Var*> *boundVars;
	List<Var*> *joinVars;
	// The current closest disjunctive parent
	Pattern* joining;
};

struct Fnarg
{
	Name name;
	Type *type;
};
Fnarg *mkFnarg(Name name, Type *type);

// Types need to persist beyond compilation,
// since they get embedded into the instruction stream.
// ASTGEN ARENA: none

struct Gram;
template<typename T> struct CfgSym;
template<typename T> struct CfgRule;

typedef enum TypeOp
{
	TypeInt = 1,
	TypeBool,
	TypeFloat,
	TypeString,
	TypeList,
	TypeFn,
	TypeTuple,
	TypeStruct,
	TypePtr,
	TypeArray,
	TypeMap,
	TypeMapIter,

	TypeDontCare,	// wildcard
	TypeName,	// placeholder for to-be-resolved typedef
	TypeSubname,	// placeholder for to-be-resolved typedef
	TypeUnresolved,	// bad TypeName or TypeSubtype

	// grammar types
	TypeSyntax,         // grammers, symbols, rules, and abstract syntax
	TypeUnion,          // eg, (C + XLambda + ?)

	// polymorphism
	TypeVar,
	TypeForall,
	TypeSubst,
} TypeOp;
struct Type
{
	TypeOp op;
	Type *left;
	Type *right;
	List<Type*> *types;
	List<Var*> *vars;
	List<Type*> *typevars;
	Name name;
	// Named types are defined local to a scope.
	Scope *declscope;
	// How much space to reserve at the front of the structure
	// for builtin struct types.  This should be 0 for tags.
	int reserved;
	// ------- Extensible Structs -------
	int extensible;
	// A list of structs which represent extensions to this struct.
	// For a tag struct, these are extensions for that tag.
	List<Type*> *extensions;  // List(Type*)
	Type *extends;     // the struct being extended (if any)
	// ------- Discriminated Structs -------
	int discriminated; // this should not be set for tags
	// List of struct types representing the contents of each
	// discrimination.  The name contained in each struct type
	// corresponds to its tag.  This list may be empty (an
	// extensible struct may start out with no tags).
	// This is only meaningful for the base struct.
	List<Type*> *tags;       // List(Type*)
	Hash<Name,List<Type*>*> *tagshash;   // mirrors 'tags'
	// There is a single inheritence relationship between tags in
	// a discriminated struct.  For tag structs, `super' is a link
	// to the parent tag struct, or the base struct if no parent
	// was specified.  This should be nil for the base struct.
	Type *super;
	Type *_supername; // ugly hack for name resolution, sorry; see typesubname()
	// --------------
	// For structs, tag structs, and tuples, `ctorargs' should
	// contain the list of argument types to the constructor (or,
	// equivalently, the argument types to the pattern-matching
	// destructor).  For structs and tuples, this is the same as
	// `types'.  For tag structs, it's the concatenation of
	// `tagof->types' and `types'.  For extension structures,
	// this is meaningless.
	List<Type*> *ctorargs;
	// Is this type fully resolved?  If this is 0, then it is
	// unknown.  In general, this flag should not be accessed
	// directly.  Use isunresolved() instead.
	int isResolved;
	// ------- Grammars ---------
	// For TypeSyntax types, this points to the underlying struct
	// representation.  There is one struct for each sym and rule
	// in a grammar family.  For example, C.expr and (C+XLambda).expr
	// use the same structrep.  See grammartyping.c for details.
	Type *structrep;
	// All attributes are stored in structrep->vars.  The total
	// size of all attributes is given by structrep->attrsize.
	int attrsize;
  // The underlying grammar (the same for all TypeSyntaxs in the same family).
	struct Gram *grammar;
	// Points to the CfgSym* or CfgRule* for this type, if it is a sym or a rule.
	union {
		struct CfgSym<AstToken>  *cfgsym;
		struct CfgRule<AstToken> *cfgrule;
	};
	// For grammar types (a TypeSyntax with super == nil, or a TypeUnion),
	// the list of grammars which are supertypes of this grammar.  Any time
	// a grammar X extends this grammar, it gets added to 'supergrammars'.
	List<Type*> *supergrammars;
	// Hack for views (used to support Type <-> TypeRep)
	Type *mirrors;
	Type *mirroredby;
	View *view;
};
Type *mkType(TypeOp op);
Type *mkTypeBool(void);
Type *mkTypeInt(void);
Type *mkTypeFloat(void);
Type *mkTypeString(void);
Type *mkTypeList(Type *left);
Type *mkTypeFn(List<Type*> *types, Type *left);
Type *mkTypeTuple(List<Type*> *types, List<Type*> *ctorargs);
Type *mkTypeStruct(Scope *declscope, Name name, Type *super);
Type *mkTypePtr(Type *left);
Type *mkTypeName(Name name, List<Type*> *types);
Type *mkTypeSubname(Type *_supername, Name name);
Type *mkTypeUnresolved(Scope *declscope, Name name);
Type *mkTypeArray(Type *left);
Type *mkTypeMap(Type *left, Type *right);
Type *mkTypeMapIter(Type *left, Type *right);
Type *mkTypeDontCare();
Type *mkTypeSyntax(Name name, Type *super, Type *structrep, struct Gram *grammar);
Type *mkTypeUnion(List<Type*> *types, struct Gram *grammar);
Type *mkTypeVar(Name name);
Type *mkTypeForall(List<Type*> *typevars, Type *left);
Type *mkTypeSubst(List<Type*> *types, List<Type*> *typevars, Type *left);
char *strTypeOp(int);

Type *typevoid(void);
Type *typedontcare(void);
Type *typebool(void);
Type *typeint(void);
Type *typefloat(void);
Type *typestring(void);
Type *typeptr(Type *left);
Type *typelist(Type *left);
Type *typearray(Type *left);
Type *typemap(Type *left, Type *right);
Type *typemapiter(Type *left, Type *right);
Type *typetupleexprs(const vector<Expr*>& exprs);
Type *typetuple(List<Type*> *types);
Type *typeunion(List<Type*> *types, struct Gram *grammar);
Type *typefnargs(List<Fnarg*> *fnargs, Type *rettype);
Type *typefn2(Type *in, Type *rettype);

// typedefs
Type *typesubname(Type *super, Name name);
Type *typeunresolved(Scope *scope, Name name);

// polymorphism
Type *typeforall(List<Type*> *typevars, Type *left);
Type *typesubst(List<Type*> *types, List<Type*> *typevars, Type *left);

List<Type*> *names2typevars(List<Name> *names);

// builtin types
Type *typeast(void);         // abstract type Ast
Type *typeaststring(void);   // abstract type for AstStrings (terminals)
Type *typeline(void);

// not actually builtin, but we know about them
Type *typeframe(void);
Type *typescope(void);
Type *typesym(void);

// ASTGEN ARENA: asyntax

int isassoclist(Type *type, Type **key, Type **val);

typedef enum StructelemOp
{
	StructelemDecl = 1,
	StructelemTag,
} StructelemOp;
struct Structelem
{
	StructelemOp op;
	Line line;
	Fnarg *decl;
	Name tagname;
	Name supername;
	List<Structelem*> *tagdecls;
};
Structelem *mkStructelemDecl(Fnarg *decl);
Structelem *mkStructelemTag(Name tagname, Name supername, List<Structelem*> *tagdecls);

typedef enum PGramOp 
{
	PGramRule = 1,
	PGramIgnore,
	PGramAssoc,
	PGramPriority,
	PGramPrefer,
	PGramShare,
	PGramView,
} PGramOp;
struct PGram
{
	PGramOp op;
	Line line;
	Name name;
	Name name2;
	Name prec;
	GTerm *left;
	List<GTerm*> *terms;
	List<Name> *names;
	List<List<Name>*> *lnames;
	int assoc;
};
PGram *mkPGramRule(Line line, Name name, GTerm *left, List<GTerm*> *terms, Name prec);
PGram *mkPGramIgnore(Line line, List<GTerm*> *terms);
PGram *mkPGramAssoc(Line line, int assoc, List<Name> *names);
PGram *mkPGramPriority(Line line, List<Name> *names);
PGram *mkPGramPrefer(Line line, List<Name> *names);
PGram *mkPGramShare(Line line, List<List<Name>*> *lnames);
PGram *mkPGramView(Line line, Name name, Name name2, List<List<Name>*> *lnames);

typedef enum GTermOp
{
	GTermName = 1,
	GTermString,
	GTermRegexp,
	GTermOpt,
	GTermStar,
	GTermPlus,
} GTermOp;
struct GTerm
{
	GTermOp op;
	Line line;
	Name name;
	char *str;
	int len;
	int flags;
};
GTerm *mkGTermName(Name name);
GTerm *mkGTermString(char *str, int len);
GTerm *mkGTermRegexp(char *str, int len, int flags);
GTerm *mkGTermOpt(Name name);
GTerm *mkGTermStar(Name name, char *str);
GTerm *mkGTermPlus(Name name, char *str);

typedef enum PTokenOp
{
	PTokenString = 1,
	PTokenExpr,
} PTokenOp;
struct PToken
{
	PTokenOp op;
	Expr *expr;
	char *str;
	int len;
	// for PTokenExpr
	Name symname;  // ::sym annotation
	Name elemname; // for lists, symname of the list element
	Name slot;     // name of the slot (for expr->op == ExprName)
	int islist;    // expression has type "list ptr Ast"
};
PToken *mkPTokenString(char *str, int len);
PToken *mkPTokenExpr(Expr *expr);

// ASTGEN ARENA: none

struct StmtGrammar;

struct Yystype
{
	Type *type;
	union {
		Stmt *stmt;
		Fnarg *fnarg;
		Expr *expr;
		ExprParse *exprparse;
		ExprMap *exprmap;
		ExprInitField *initfield;
		Structelem *structelem;
		Clause *clause;
		Pattern *pattern;
		struct {
			Name name;
			int len;
			int flags;
		};
		double f;
		int i;
		GTerm *gterm;
		PGram *pgram;
		int assoc;
		StmtFndef *stmtfndef;
		StmtStruct *stmtstruct;
		StmtGrammar *stmtgrammar;
		vector<Expr*> *exprs;
		vector<ExprMap*> *exprmaps;
		vector<ExprInitField*> *initfields;
		List<Expr*> *exprlist;
		List<Stmt*> *stmtlist;
		List<Name> *namelist;
		List<GTerm*> *gtermlist;
		List<Structelem*> *structelemlist;
		List<Fnarg*> *fnarglist;
		List<Clause*> *clauselist;
		List<Pattern*> *patternlist;
		List<PGram*> *pgramlist;
		List<Type*> *typelist;
		List<PToken*> *ptokenlist;
		List<List<Name>*> *namelistlist;
	};
};
#define YYSTYPE Yystype

int yylex(Yystype*, Line*);
void yyerror(char*);
int yyparse(void);
extern int inbacktick;
extern int ingrammar;
extern int intermname;

char*		cescapestring(char *s, int len, int q);
char*		cunescapestring(char *s, int len);
int		parsechar(char *s, uint64 *value, char **end);
int		parsefloat(Line line, char *s, long double *value, Type **type);
int		parseinteger(Line line, char *s, uint64 *value, Type **type);

void lexinit(char*);
int  isinteractive(void);
int  pushinput(char*, char*);
void popinput(void);
void doinclude(char*);
void dorequire(Line, Name, int);
void loadextension(char*);

struct PatternXForm
{
	Codebuf *codebuf;
	int off;            // offset into codebuf->bp; the Ast* to patch
	View *view;
	Line line;
};
extern List<PatternXForm*> *queuedpatternxforms;
void astviewxform(Thread*, PatternXForm*);

union Datum
{
	int i;
	double f;
	void *p;
};

struct Thread
{
	// Instruction stream contains
	// function pointers mixed with immediate data.
	union {
		Code **c;
		int *i;
		double *f;
		void **p;
		uchar *b;
	} pc_;
	void *fp_;
	void *gp_;
	jmp_buf stop_;
	int depth_;
	int rundepth_;
	int upcall_;     // depth of the current upcall
	List<void*> *upcalls_;  // stack of upcall fps
	
	void *loc();
	
};

struct LabelStub
{
	Label *prev;
	Label *next;
};

struct Label
{
	Label *prev;
	Label *next;
	Name name;
	Stmt *stmt;
	Stmt *used;
	char *val;
	List<int*> *uses;
};

struct Codebuf
{
	uchar *bp;
	uchar *ep;
	union {
		uchar *p;
		int *ip;
		double *fp;
		void **vp;
	};
	LabelStub labels;
	int live;
	Type *rettype;
};

Codebuf *mkcodebuf(void); // noauto
void compilestmt(Compile*, Stmt*);
void compileexpr(Compile*, Expr*);
void compilepattern(Compile*, Pattern*, Loc*, Label*, int *canfail, int istmp);
void run(Thread*);
void callfn(Thread*, Fn *fn, void *args);
void callglobalfn(Thread*, Name name, void *args, int argsize);

struct Frame
{
	List<Var*> *vars;        // local to this frame
	List<Var*> *allvars;     // vars + escaped vars
	Frame *outer;
	int size;
	int argsize;
	int escsize;
	int off;
	int tmpoff;
	int depth;
};

extern Frame *globalframe;
extern Scope *globalscope;
extern Scope *realglobalscope;

Frame *mkFrame(void);
void* mkfnframe(Fn *fn); // noauto
void freefnframe(void*);

struct Scope
{
	Frame *frame;      // the containing frame
	List<Var*> *end;         // the first var outside this scope (a pointer into frame->vars)
	List<Var*> *start;       // the innermost var in this scope (if nil, use frame->vars.hd)
	Scope *outer;      // outer scope (this might cross frame boundaries!)
	List<Scope*> *imported;    // scopes imported via "require"
	List<Type*> *grams;       // types of grammars visible in this scope
	List<Var*> *canonicals;	// active canonicalizers
};
Scope *mkScope(Frame *frame, Scope *outer, List<Var*> *end);

void printframe(Frame*, int);
void printscope(Scope*, int, int);

struct Var
{
	Name name;
	Name mangled;
	Type *type;
	Frame *f;
	int off;
	Label *label;
	int isconst;
	int escapes;
	int argoff;
	int istype;
	int isdefined; // if this is a function or attribute, does it have a definition yet?
	Var *attrfn;      // this is an attribute Var if (attrfn != nil || isbuiltinattr)
	int isbuiltinattr;
	Type *host;
	Scope *declscope; // for attributes, the scope this is declared in
};
Var *mkVar(Name name, Type *type);

#include "compile.h"

struct Fn
{
	Name name;	// just a hint
	Type *type;
	List<Fnarg*> *fnargs;
	Type *rettype;
	List<Stmt*> *stmts;
	Codebuf *code;
	Frame *frame;
	Compile *cs;
	Typecheck *tc;
	Fn *outer;
	Fn *link;
	void (*builtin)(void*);
	Line line;
	
	// Run-time profiling.
	Hash<Fn*,uintptr> *callouts;	// count of times has called other Fns
	int calls;	// count of times has been called
	int closures;	// number of closures allocated in this fn
};
Fn *mkfn(Line line, Name name, List<Fnarg*> *fnargs, Type *rettype, List<Stmt*> *stmts);  /* noauto */
void compilefn(Line line, Fn *fn);
extern List<Fn*> *allfns;

// Deep syntax copying
Fn *copyfn(const Fn*);
Stmt *copystmt(const Stmt*);
Expr *copyexpr(const Expr*);
Clause *copyclause(const Clause*);
Pattern *copypattern(const Pattern*);
PToken *copyptoken(const PToken*);

// Type checking
enum
{
	Value = 1<<0,		// Expr being used for its value
	LValue = 1<<1,		// Expr being used for its lvalue
	BindValue = 1<<2,	// Expr must be a name
};

#include "typecheck.h"
void typecheckstmt(Typecheck*, Stmt*);
void typecheckexpr(Typecheck*, Expr*, int);

// This expression cannot be assigned to.
void noassign(Typecheck*, Expr*, int);

// Using this expression only for side effects is a mistake.
void nosideeffect(Typecheck*, Expr*, int);

void typecheckstmtcase(Typecheck*, Stmt*);
void typecheckstmtforlist(Typecheck*, Stmt*);
void typecheckpattern(Typecheck*, Pattern*, Type*);

void typecheckexprdestructure(Typecheck*, Expr*, int);
void typecheckexprdestructtype(Typecheck*, Expr*, int);
void typecheckexprrestructure(Typecheck*, Expr*, int);
void typecheckexprparse(Typecheck*, Expr*, int);
void typecheckexprparsestring(Typecheck*, Expr*, int);
void typecheckexprastparent(Typecheck*, Expr*, int);


void translatemkscope(Typecheck*, Expr*);
void translatescopemerge(Typecheck*, Expr*);

void pushscope(Typecheck*);
void popscope(Typecheck*);
Label *mklabel(Typecheck*);  // noauto
void pushbreak(Typecheck*, Label*);
void popbreak(Typecheck*, Label*);
void pushcontinue(Typecheck*, Label*);
void popcontinue(Typecheck*, Label*);
Label *uselabel(Typecheck*, Name, Stmt*);
Label *setlabel(Typecheck*, Name, Stmt*);
void pushflow(Typecheck*);
void popflow(Typecheck*);
void addtyperefinement(Typecheck*, Expr*, Type*);

void addstructtag(Type*, Type*);
Var *addstructattr(Name, Type*, Type *host, int isbuiltin);
void fixglobalattrs(void);

Var *lookup(Scope*, Name, int local);
Type *lookuptype(Scope*, Name, int local);
Var  *lookupstruct(Type*, Name, Scope*, Type**);
Type *lookuptag(Type*, Name, Scope*);
Var  *lookupattr(Type*, Name, Scope*, int dosuper);
Type* lookuptagowner(Type*);
Type* rootstruct(Type*);

int checkunresolved(Line, Type*);
Type *resolvenamedtype(Typecheck*, Line, Type*);
Type *_maptypechildren(void*, Type*, Type*(*)(void*, Type*));

template<typename T>
Type*
maptypechildren(T *arg, Type *t, Type *(*xform)(T*, Type*))
{
	return _maptypechildren((void*)arg, t, (Type*(*)(void*,Type*))xform);
}

Var *escapev(Var*);
void escapef(Frame*);
int declareexpr(Typecheck*, Expr*, Type*);
Var *declare(Typecheck*, Line, Name, Type*);
Var *declareunique(Typecheck*, Line, Name, Type*);
Var *declaretype(Typecheck*, Line, Name, Type**);
Fn *typecheckfn(Typecheck*, Line, Fn*, Var*, List<Name>*, int);
Type *typecheckstruct(Typecheck*, Line, Name, List<Type*>*, List<Structelem*>*, Type*, Type*, int, int);
void dotop(Stmt*);
void showprompt(void);


int sizeoftype(Type*);
int structdataoffset(Type*);
int tagtypeoffset(Type*);
int extptroffset(Type*);

int isabstracttype(Type*);
int issubtype(Type*, Type*);
int issubtypesyntax(Type*, Type*);
int issubtypegrammar(Type*, Type*);
int isverticalsubtype(Type*, Type*);
Type* subtypejoin(Type*, Type*);
int shouldcompletetype(Type*, Type*);
int isunresolved(Type*, Type**);
int isptrast(Type*);

Type *applytypesubst(Type*, List<Type*>*, List<Type*>*);
int canunifytypes(Line, Type*, Type*, List<Type*>*, List<Type*>*);

Type* gramsupertype(Type*);
Type* symsupertype(Type*);
Type* rulesupertype(Type*);
Type* replacegramtype(Type*, Type*);

Type* gramroottype(struct Gram*);
Type* gramwildtype(struct Gram*);

Type* resolveparsetype(Typecheck*, Type*);
Type* resolvegramsubtype(Typecheck*, Type*, Name);
Type* resolvedefaultxgram(Typecheck*, Type*);
Type* resolvegramwildcard(Line, Name);
Type* joingrammartypes(Type*, Type*);
void adduniongrammar(Type*, struct Gram*);

enum 
{
	LocGlobal,
	LocGlobalIndirect,
	LocFrame,
	LocFrameIndirect,
	LocEscape,
	LocEscapeIndirect,
	LocIndirect = 1,
};
struct Loc
{
	int mode;
	int off;
	int off1;
	int depth;
	Type *type;
};

Code xstop;
void emitp(Compile*, void*);
void emitC(Compile*, Line, Code*);
void emiti(Compile*, int);
void emitloadptr(Compile*, Line, Loc*, void*);

int typefmt(Fmt*);
int exprfmt(Fmt*);
int stmtfmt(Fmt*);
int patfmt(Fmt*);
int locfmt(Fmt*);
void disasm(uchar*, uchar*);
char *disasmfnname(void*);
void patch(Compile*);

int pprintf(Fmt*, char*, ...);
void pprintstmt(Fmt*, Stmt*);
void pprintexpr(Fmt*, Expr*);
void pprintpattern(Fmt*, Pattern*);

Type *typefn(List<Type*>*, Type*);

void emitmov(Compile*, Line, Loc*, Loc*);
Fn *mkbuiltinfn(Name name, void(*fn)(void*), Type*); /* noauto */
void installbuiltin(char*, void(*)(void*), Type*);
void installbuiltindata(char*, void*, Type*);
void installbuiltintype(char*, Type*);
void builtininit(int, char**);

// Attribute data blocks
struct AttrBlock
{
	void *attrs;
	int size;
};

// Builtin line structure
struct ZLine
{
	// hidden variables - must be first

	// struct header fields - must be before the shared variables
	// (there are none, because this is a plain struct)

	// variables shared with runtime - must be last
	// must match builtin.c
	Name file;
	int line;
};

// Extensible grammar parsed structures
typedef enum
{
	AstString = 1,
	AstRule,
	AstSlot,
	AstListSlot,
	AstMerge,
} AstOp;
struct Ast
{
	// hidden variables - must be first
	AstOp op;
	struct CfgSym<AstToken> *sym;
	Ast **_right;      // for AstRule, AstMerge; do not access directly; use astright.
	int nright;
	Ast *sep;          // for separated lists only (a [name sep] grammar term)
	Name text;         // AstString: matched text
	                   // Ast*Slot:  slot name (as a printed expression for \() slots)
	                   // otherwise: cached asttoa()
	uchar hasslots;    // does this Ast have slots?
	uchar listkind;
	
	AttrBlock gram_attrs; // attributes for G, G.sym, and G.sym.rule.
	AttrBlock sym_attrs;  // need one for each because they are extended individually
	AttrBlock rule_attrs;

	int traceid;

	// struct header fields - must be before the shared variables
	ExtStruct *extensions;

	// This is a TypeSyntax.  It is the most specific type which can describe
	// the entire syntax tree rooted at this node.  More specifically, as a
	// case analysis on Ast->op:
	//   AstRule: the most specific grammar rule which could be used to parse
	//            this abstract syntax tree (a least-upper-bound)
	//  AstMerge: the symbol type, in the form G.sym."merge", where G the join
	//            of the grammar types of both branches
	//  Ast*Slot: this slot type, as declared or inferred (could have the form
	//            G, G.sym, or G.sym.rule)
	// AstString: the special typeaststring() type
	Type *_specifictype;

	// variables shared with runtime - must be last
	// must match builtin.c
	struct CfgRule<AstToken> *rule; // exposed as ptr(dontcare)
	Ast *_parent;
	Ast *copiedfrom;
	int isshared;    // is this a shared Ast?
	Name symname;      // CfgSym::name may be obfuscated (see grammartyping.c),
	                   // but this is not
	ZLine zline;
	// Tag fields go here, but all tag structs for Ast.sym are
	// initially empty.  So, sizeoftype(Ast.sym) == sizeof(struct Ast)
	// for all 'sym'.
};

int astfmt(Fmt*);
Array *astsplit(Ast*);
Name asttoa(Ast*);
Ast *astincref(Ast*);
void astdecref(Ast*);
List<Ast*> *asttolist(Ast*);
ListNode *asttolistnode(Ast*);
Ast *astjoin(Ast*, int, Ast**);
Ast *astdeepcopy(Ast*);
void astprevnext(Ast*, Ast**, Ast**);
Ast *astlastkid(Ast*);
Ast *astrealparent(Ast*);
Ast *asttypedparent(Ast*, Type*);
Ast *astright(Ast*, int);
Ast *astparent(Ast*);
Ast *astcopy(Ast*, int);

Type* addgrammar(Typecheck*, StmtGrammar*);
int grammarfmt(Fmt*);

Ast* parsebuiltin(Type *start, char *text, Name file);
int parsepattern(ExprParse*, int infertypes);
int samestructure(Ast*, Ast*);
int destructure(List<Var*>*, Ast*, Ast*, int n, void ***slots);
int destructtype(Ast*, Name, Name);
Ast *_restructure(Ast*, int n, void **slots, Ast*(*)(Ast*,void*), void*);
Ast *restructslot(Type*, Ast*);
void astcollectslots(Ast*, int*, Ast***);
int isambiguous(Ast*);

template<typename T>
Ast*
restructure(Ast *ast, int n, void **slots, Ast*(*fn)(Ast*, T*), T *arg)
{
	return _restructure(ast, n, slots, (Ast*(*)(Ast*,void*))fn, arg);
}

void addrestructure(Line, Name);

int isastlistsymname(Name, Name*);
int isastlisttype(Type*, Type**);
int isastoptionsymname(Name, Name*);
int isastoptiontype(Type*, Type**);
int isreservedsymname(Name);

// A view is a custom pattern-matcher for some internal representation.
// Say we have a view:
//
//   view Type on TypeRep where {
//      pattern = abtype
//      slotin  = TypeRep abtype
//      slotout = abtype  Type
//      xform = abtype2typerep
//   }
//
// A pattern like Type{\T **} is handled in the following way:
// 1. slots which have type 'ptr TypeRep' shift [abtype]
// 2. the pattern is parsed using 'abtype'
// 3. if a slot is inferred to be '::abtype', it it typed to 'ptr Type'
// 3. the parsed ast is transformed by 'abtype2typerep'
// 4. the final ast is used directly by all de/restructuring operations.
//
struct View
{
	Type *viewtype;    // type which invokes this view
	Type *reptype;     // internal representation
	Type *parseme;     // exposed parsing interface
	Hash<Name,List<CfgSym<AstToken>*>*> *slotins;
	Hash<Name,Name> *slotouts;
	Name xformname;    // the function which is transforms 'parseme' --> 'reptype'
	Var *xform;
	Name postxformname;// an optional function which cleans up a restructured view ast
	Var *postxform;
};

// Data layout for structs
// Angle <> brackets denote required data, and
// square [] brackets denote optional data.
//
// With this layout, Zeta can cast struct pointers and
// access tag fields with minimal pointer dereferencing.
//
// {
//    [reserved]      // block of data of size Type.reserved
//    [ExtStruct*]    // list of extensions, if extensible (allocated on-demand)
//    [Type*]         // type of the tag, if discriminated
//    <base fields>   // in order of declaration, the base struct fields
//      :             // ... and down the heirarchy ...
//    [super fields]  // the tag->super's fields
//    <tag fields>    // the tag fields
// }
//
// Data layout for extension structs
//
// {
//    <ExtStruct*>    // next in the chain
//    <Type*>         // type of the extension struct
//    <ext fields>    // in order of declaration, the extension fields
// }

void *newstruct(Type*);
void copystruct(void*, const void*, Type*, Type*);
void copyxstructs(ExtStruct*, const ExtStruct*, Type*, Type*);

struct ExtStruct
{
	ExtStruct *next;
	Type *type;
	char v[1];
};


// Builtin data structures

Array *newarray(int n, int sizeof1);

struct Array
{
	int n;
	int sizeof1;
	void *v;
};

struct ListNode
{
	ListNode *next;
	char v[1];
};

ListNode *listtolistnode(List<void*>*);
List<void*> *listnodetolist(Arena *a, ListNode*);

struct AssocList
{
	AssocList *next;
	void *k;
	void *v;
};

struct Map
{
	Hash<void*,void*> *h;
	int sizeof1;
};

struct MapIter
{
	Hashiter<void*,void*> iter;
	Hashkv<void*,void*> kv;
	int sizeof1;
	void *k;
};

enum
{
	FrameOffsetPC = 0,
	FrameOffsetFP = sizeof(void*),
	FrameOffsetFN = 2*sizeof(void*),
	FrameOffsetEP = 3*sizeof(void*),	// not always present
	FrameReserved = 3*sizeof(void*)
};

void checkdisasm(void);

void *galloc(int);
extern int memtot, frametot;

void profilecall(Fn *caller, Fn *callee);
void profileclosure(Fn *creator, Fn *closure);
void printprofile();

void *var2ptr(Thread*, Var*);

int file2num(char*);
Name num2file(int);

void	dumpstack(Thread*);
extern Thread *running;

#define NoParentYet ((Ast*)-1)

void *computeattr(Ast*, Var*);

// Debugger
extern int debugfd;
void debuginit(void);
void debugbreak(void);
void tracemkast(Ast *ast);
void traceenterattr(Ast *ast, Name name, Type *host);
void traceleaveattr();
void tracereadattr(Ast *ast, Name name, Type *host);

