// Grammar extension.
// ==================================================================
//
// This is one of the core ideas that makes zeta very different from
// other poorly-implemented scripting languages.
//
// EXAMPLES (see also test/xgram*.zeta)
//
//	grammar {
//		%ignore /[ \t\n]+/;
//		expr: expr "+" expr;
//		expr: expr "-" expr;
//		expr: expr "*" expr;
//		expr: "(" expr ")"
//	}
//
//	if(e ~ expr{\a * 2})
//		return `expr{\a::expr + \a};
//
//	if(e ~ expr.merge{\a, \b})
//		xxx;
//
//	grammar {
//		expropt: expr?
//	}
//
//	e: ptr Ast.expr?;
//	e1: ptr Ast.expr;
//	if(e ~ expr?{\e1})
//		xxx
//
//	e1 = option(e);	// Want something better here; option is too vague.
//
//	grammar {
//		exprlist: expr+
//		exprlistopt: expr*
//		exprcommalist: expr[,]+
//		exprcommalistopt: expr[,]*
//	}
//
//	el := `expr+{1 2};
//	elo := `expr*{};
//	ecl := `expr[,]+{1, 2, 3};
//	eclo := `expr[,]*{};
//
// IMPLEMENTATION
//
//	XXX something about Asts and grammars in general here
//	XXX shorthands
//
//	Options
//
//	The shorthand expr? is defined as:
//
//		expr?: /*empty*/
//		expr?: expr
//
//	At the moment expr? is treated as a first-class symbol as far as matching:
//	you can have ptr Ast.expr?, etc.  In the future I'd like to change this so that
//	you only have ptr Ast.expr, and matching a ptr Ast.expr? slot simply has the
//	possibility of setting the slot to nil.
//
//	Lists
//
//	The shorthands expr*, expr+, expr[,]*, and expr[,]+ are defined as:
//
//		expr@: expr
//		expr@: expr@ expr
//
//		expr+: expr@
//		expr*: expr@
//		expr*: /*empty*/
//
//		expr[,]@: expr
//		expr[,]@: expr[,]@ "," expr
//
//		expr[,]+: expr[,]@
//		expr[,]*: expr[,]@
//		expr[,]*: /*empty*/
//
//	All of the '@' symbols are hidden from zeta programs.
//
//	Parsing Patterns
//
//	expr{\a * 2} is called a parsing pattern.  Patterns are used both for 
//	destructuring (taking apart) and restructuring (putting back together)
//	abstract syntax.  The \a in the pattern is called a slot.  It stands
//	for an arbitrary piece of abstract syntax (subject to some contraints).
//	Parsing expr{\a * 2} produces an Ast with slots in place of each 
//	escaped expression like \a.  This Ast can be passed:
// 
//		- to destructure to matching against another Ast x,
//		  returning the sub-Asts in x that corresponded to the slots
// 
//		- to restructure along with a set of values, 
//		  returning a new Ast with the values substituted for the slots
//
//	Each slot is statically determined to be one particular symbol type
//	(like "expr" or "expr[,]+").  There are multiple cues used to determine
//	the symbol type.  They are:
//
//		1. The type of the expression used in the slot:
//			- if e is ptr Ast.expr, then \e is an "expr".
//			- if e is list ptr Ast.expr, then \e is some list of expr,
//		      like "expr+" or "expr[,]+".
//
//		2. The optional ::suffix on the slot expression:
//			- \e::expr is an "expr",
//			  and e is a ptr Ast.expr (or just ptr Ast)
//			- \l::expr[,]+ is an "expr[,]+",
//			  and l is a list ptr Ast.expr (or just list ptr Ast)
//
//		3. The simplest possible choice (shallowest parse) given the
//		   constraints.  For example, in expr{\a * 2}, \a could be
//		   an expr as in "expr => expr * expr", or it could be a number
//		   as in "expr => (expr => number) * expr", but the latter
//		   is a more complex parse and is ignored.
//
//	Rule #3 requires building all possible parses that match the first
//	two constraints.  A GLR parser does this for us, but we must still 
//	give it some kind of type for \a to work with.  Specifically, we 
//	temporarily give \a one of the following wildcard types:
//
//		.any   = any Ast in the grammar
//		@.any  = any internal list
//		@expr  = any list of expr (e.g., expr+, expr[,]+)
//
//	These are defined as ordinary non-terminals in the GLR grammar,
//	though the names are not accessible from zeta programs.  They are
//	defined by creating new grammar rules, for each symbol expr:
//
//		expr: .any
//		expr@: @.any
//		expr[,]@: @.any
//		expr@: @expr
//		expr[,]@: @expr
//
//	(expr@ was defined above, in the discussion of lists).
//
//	The static type system does not guarantee that list types are correct
//	in restructuring expressions.  For example, if l: list ptr Ast, then
//	in `expr{f(\l)}, l will be deduced to be a "expr[,]+".  At run time,
//	the interpreter panics if list l is empty.
//
//	TODO
//	summary of modularity?
//	there are some Type-based comments in grammartyping.c

#include "a.h"
#include "cfg.h"
#include "slr.h"
#include "nfa.h"
#include "dfa.h"
#include "glr.h"

typedef struct AstToken AstToken;
typedef struct Gram Gram;

struct AstToken
{
	Ast *ast;
	int tokno;
	int tokno1;
	int tokno2;
	ZLine zline;
};

// special CfgSym* returned by DFA to indicate
// "skip over this text".
#define Ignore ((CfgSym<AstToken>*)-1)

struct Gram
{
	CfgGram<AstToken> cfg;

	Name name;
	Type *roottype;      // the type of the root grammar
	Type *wildroottype;  // the type (root + ?)

	Type *asttype;       // the type of this grammar module (TypeAst)

	int dfaprec;         // current dfa precedence (decreases)
	CfgSym<AstToken> *any;         // cfglooksym(gram, ".any")
	CfgSym<AstToken> *anylist;     // cfglooksym(gram, "@.any")
	CfgSym<AstToken> *eof;         // cfglooksym(gram, "EOF")
	CfgSym<AstToken> *error;       // cfglooksym(gram, "error")
	CfgSym<AstToken> *slot;        // cfglooksym(gram, "slot")
	CfgSym<AstToken> *merge;       // cfglooksym(gram, "merge")

	DFA<CfgSym<AstToken>*> *lexer;          // DFA used for lexing.
	Hash<Name,CfgSym<AstToken>*> *literals;
	Hash<Name,CfgSym<AstToken>*> *regexps;
	Hash<CfgSym<AstToken>*, intptr> *sharedsyms;    // symbols which are maximally shared
	Hash<CfgRule<AstToken>*, intptr> *sharedrules;   // rules which are maximally shared
	List<Name> *ignores;       // ignored symbol names ("L.ingore" or "R.ignore")

	List<Type*> *wildtypes;     // wildcard types based on this grammar
	List<Type*> *symtypes;      // all symbol and rule types under this grammar
	List<Type*> *ruletypes;
};

// grammarbuild.c
int astlistkind(CfgSym<AstToken>*);
void astfixlistparents(Ast*);

Ast* mkaststring(CfgSym<AstToken>*, Name);
Ast* mkastrule(CfgRule<AstToken>*, Ast**);
Ast* mkastmerge(Ast*, Ast*);
Ast* mkastslot(Type*, Name);
Ast* mkastlistslot(Type*, Name);
Ast* mkastlist(Ast*, List<Ast*>*);

Ast* findsharedast(CfgRule<AstToken>*, Ast**);
void memosharedast(Ast*);

extern Line GramLine;
Type *findgrammar(Name);

// grammarparse.c
int action(CfgGram<AstToken>*, CfgRule<AstToken>*, AstToken*, AstToken*);

// grammartyping.c
Type* addgramtype(Gram*, Name, Scope*);
Type* addsymtype(Type *gramtype, CfgSym<AstToken>*, Name);
Type* addruletype(Type *symtype, CfgRule<AstToken>*, Name);
void addbuiltingramtags(Type*, void*, void*);
void resolvesyntaxsubtype(Name, Type *super, Type *canon);

static inline Type* _cfgsymtype(CfgSym<AstToken> *s) {
	return (Type*)s->_origin;
}

static inline Type* _cfgruletype(CfgRule<AstToken> *r) {
	return (Type*)r->_origin;
}

View* mkgramview(Type*, Name, Type*);

// grammarstruct.c
void astremovecopiedfrom(Ast*);

struct StmtGrammar : public Stmt
{
	StmtGrammar(Name name, List<Name> *extends, List<PGram*> *grams) :
		Stmt(), name_(name), extends_(extends), grams_(grams),
		flags_(0) {}
	
	Name name_;
	List<Name> *extends_;
	List<PGram*> *grams_;
	
	int flags_;
	
	virtual void compile(Compile*);
	virtual void typecheck(Typecheck*);
	virtual void pprint(Fmt*);
};

struct StmtCanonicalize : public Stmt
{
	StmtCanonicalize(List<Name> *names) :
		Stmt(), names_(names) {}
	
	List<Name> *names_;
	
	virtual void typecheck(Typecheck*);
	virtual void compile(Compile *cs) {}
};
