#include "grammar.h"

// see comment in grammarbuild.c
#define lookuptag(a, b)   lookuptag(a, b, nil)

// For debugging parse patterns: print structure of every pattern parsed.
DEFINE_int32(parsedebug, 0, "Parse debug level");
DEFINE_int32(parsemergedebug, 0, "Parse merge debug level");

static int validateast(Ast*, Type*, int);
static int validateastrule(CfgRule<AstToken>*, Type*);

// ========================================================
// Parsing of regular text (not patterns)

// Two different parses, a and b, have been found for 
// the same segment of text.  Choose which to use,
// leaving the answer in atok.
static void
glrmerge(CfgSym<AstToken> *sym, AstToken *atok, AstToken *btok)
{
	Ast *a, *b;
	CfgPrec<AstToken> *aprec, *bprec;
	int r, ta, tb;
	int debug=0;
	
	a = atok->ast;
	b = btok->ast;
	
	ta = -1;
	tb = -2;

	// fprint(2, "merge? %R || %R\n", a->rule, b->rule);

	// Can get these when some productions can yield empty strings.
	// For example:
	// 	grammar {
	// 		S: A S "b";
	// 		S: "x";
	// 		A: ;
	// 	}
	// 	`S{xbbbbb};
	if(a->rule && a->rule == b->rule && a->rule->nright == 0)
		goto A_wins;

	aprec = a->rule ? a->rule->prec : nil;
	bprec = b->rule ? b->rule->prec : nil;

	// Rule preference resolution.
	if(aprec && bprec && aprec->kind == CfgPrecPrefer && bprec->kind == aprec->kind){
		r = cfgcompareprec(a->rule->prec, b->rule->prec);
		if(r != 0)
			debug = 1;
		if(r > 0)
			goto A_wins;
		if(r < 0)
			goto B_wins;
	}

	// Rule priority resolution is done in the action rule
	// (because the merge-node parent actually does the resolution).

	// Associativity resolution.
	if(aprec && aprec->assoc && aprec == bprec){
		ta = atok->tokno1;
		tb = btok->tokno1;
		if(ta && ta == tb){
			ta = atok->tokno2;
			tb = btok->tokno2;
		}
		if(ta && ta != tb){
			debug = 2;
			if(aprec->assoc == CfgAssocLeft){
				if(ta < tb)
					goto B_wins;
				goto A_wins;
			}
			if(aprec->assoc == CfgAssocRight){
				if(ta > tb)
					goto B_wins;
				goto A_wins;
			}
		}
	}

	// Ambiguous.
	if(FLAGS_parsemergedebug)
		fprint(2,"---- Merge ----\n%R\n%R\n", a->rule, b->rule);
	atok->ast = mkastmerge(a, b);
	return;

A_wins:
	if(FLAGS_parsemergedebug){
		if(debug == 1)
			fprint(2,"---- A wins (prefer) ----\n%R\n%R\n", a->rule, b->rule);
		else if(debug == 2)
			fprint(2,"---- A wins (assoc) ----\n[%d] %#A\n[%d] %#A\n", ta, a, tb, b);
		else
			fprint(2,"---- A wins ----\n%R\n%R\n", a->rule, b->rule);
	}
	return;
B_wins:
	if(FLAGS_parsemergedebug){
		if(debug == 1)
			fprint(2,"---- B wins (prefer) ----\n%R\n%R\n", a->rule, b->rule);
		else if(debug == 2)
			fprint(2,"---- B wins (assoc) ----\n[%d] %#A\n[%d] %#A\n", ta, a, tb, b);
		else
			fprint(2,"---- B wins ----\n%R\n%R\n", a->rule, b->rule);
	}
	*atok = *btok;
	return;
}

static int
isempty(Ast *a)
{
	int i;

	if(a->op != AstRule)
		return 0;
	for(i=0; i<a->nright; i++)
		if(!isempty(astright(a, i)))
			return 0;
	return 1;
}

// If entire subtree bottoms out in a sym, return it.
static Ast*
just1(Ast *a, CfgSym<AstToken> *sym)
{
	int i;
	Ast *nonempty;

	if(a->sym == sym)
		return a;
	nonempty = nil;
	for(i=0; i<a->nright; i++){
		if(isempty(astright(a, i)))
			continue;
		if(nonempty)
			return nil;
		nonempty = astright(a, i);
	}
	if(nonempty)
		return just1(nonempty, sym);
	return nil;
}

static Ast* unmergeprec(Ast*, CfgRule<AstToken>*, CfgPrec<AstToken>*);

// Generic grammar production action.  
// Build up a new Ast with the given rule and
// right-hand-side Ast variables.
int
action(CfgGram<AstToken> *cfg, CfgRule<AstToken> *r, AstToken *out, AstToken *in)
{
	int i, nright;
	Ast *ast, *a, **right;
	AstToken *t;
	Gram *gram;

	right = nil;
	gram = (Gram*)cfg;

	right = (Ast**)emalloc(r->nsavedright * sizeof right[0]);
	nright = 0;
	for(i=0; i<r->nright; i++){
		a = in[i].ast;
		if(r->prec && r->prec->kind == CfgPrecPriority){
			// Implement operator precedence the way people
			// typically factor grammars: if the rule has a precedence,
			// it cannot refer to subrules that have lower precedence.
			// This is like rewriting:
			//	expr: expr "+" expr
			//	expr: expr "*" expr
			//	expr: number
			// into
			//	expr1: expr1 "+" expr2
			//	expr2: expr2 "*" expr3
			//	expr3: number
			// like the C committee did.
			// Note that it is *not* exactly what LALR(1) parsers like yacc do --
			// they apply this rewrite only when there is a shift/reduce conflict,
			// while this code applies them all the time.
			if((a = unmergeprec(a, r, r->prec)) == nil)
				return -1;
		}
		if(strncmp("L.", r->right[i]->name, 2) != 0)
			right[nright++] = a;
	}

if(r->nsavedright != nright) { fprint(2, "oops: %d %d; %R\n", r->nsavedright, nright, r); }
	assert(r->nsavedright == nright);

	ast = mkastrule(r, right);
	free(right);

	out->tokno = 0;
	out->tokno1 = 0;
	out->tokno2 = 0;
	out->zline.file = nil;
	for(i=0; i<r->nright; i++){
		t = in+i;

		if(!out->zline.file)
			out->zline = t->zline;

		if(out->tokno == 0)
			out->tokno = t->tokno;

		// These toknos mark the first two appearances of
		// the lhs symbol in the rhs.  They are used to
		// disambiguate by associativity in glrmerge.
		if(t->ast->sym == r->left){
			if(out->tokno1 == 0)
				out->tokno1 = t->tokno;
			else if(out->tokno2 == 0)
				out->tokno2 = t->tokno;
		}
	}
	ast->zline = out->zline;
	out->ast = ast;
	if(FLAGS_parsedebug == 1)
		fprint(2, "action: %d/%d/%d %#A\n", out->tokno, out->tokno1, out->tokno2, ast);
	return 1;
}

// If a has precedence < minprec, return nil.
// Else return a.
// If a is a merged node, toss the branches with low precedences.
static Ast*
unmergeprec(Ast *a, CfgRule<AstToken> *r, CfgPrec<AstToken> *minprec)
{
	Ast *a0, *a1;
	CfgPrec<AstToken> *prec;

	assert(minprec == r->prec);

	if(a == nil)
		return nil;
	switch(a->op){
	default:
		return a;
	case AstRule:
		// cannot have a subrule with prec < minprec
		prec = a->rule->prec;
		if(prec && prec->kind == minprec->kind && cfgcompareprec(prec, minprec) < 0){
			if(FLAGS_parsemergedebug)
				fprint(2,"---- Discard (priority) ----\nast: %R\nkid: %#A\n", r, a);
			return nil;
		}
		if(FLAGS_parsemergedebug > 1)
			fprint(2,"---- Keep (priority) ----\nast: %R\nkid: %#A\n", r, a);
		return a;
	case AstMerge:
		a0 = astright(a, 0);
		a1 = astright(a, 1);
		if(FLAGS_parsemergedebug)
			fprint(2,"---- Check Merged ----\n%R\n%R\n", a0->rule, a1->rule);
		a0 = unmergeprec(astright(a, 0), r, minprec);
		a1 = unmergeprec(astright(a, 1), r, minprec);
		if(a0 == nil)
			return a1;
		if(a1 == nil)
			return a0;
		a->_right[0] = a0;
		a->_right[1] = a1;
		if(FLAGS_parsemergedebug)
			fprint(2,"---- Still Merged ----\n%R\n%R\n", a0->rule, a1->rule);
		return a;
	}
}


// SLR parser error callback: die.
void
slrerror(CfgGram<void> *g, SlrParse *p, char *msg)
{
	panic("slrerror: %s", msg);
}

static void glrlexsharp(GlrParse<AstToken> *parse, char **pp);

// GLR lexer callback.  Should get the next token and
// return a list of GlrToken<AstToken>* representing the various
// ways to interpret that token.  Each GlrToken<AstToken> is a 
// CfgSym<AstToken>* paired with its corresponding Ast*.
// The state, in vstate, is a char** pointing at the
// current input location (itself a char*).
static List<GlrToken<AstToken>*>*
glrlexstr(GlrParse<AstToken> *parse, CfgGram<AstToken> *cfg, char **pp)
{
	char *p, *op;
	Gram *gram;
	GlrToken<AstToken> *t;
	AstToken *a;
	int len, sol;
	char *s;
	CfgSym<AstToken> *sym;
	static int tokno;
	
	gram = (Gram*)cfg;

	if(*pp == nil)
		return list1(nil, glrnewtoken(parse, gram->eof));
	if(parse->col == 1)
		glrlexsharp(parse, pp);

	List<CfgSym<AstToken>*> *l;
	do{
		p = *pp;
		if(p == nil || *p == 0)
			return list1(nil, glrnewtoken(parse, gram->eof));
		op = p;
		l = rundfa(gram->lexer, p, &p);
		*pp = p;
		if(p == op || l == nil)
			return ZList;
		sol = 0;
		for(s=op; s<p; s++){
			if(*s == '\n'){
				parse->line++;
				parse->col = 1;
				sol = 1;
			}
			else{
				parse->col++;
				if(sol && *s != ' ' && *s != '\t')
					sol = 0;
			}
		}
		if(sol)
			glrlexsharp(parse, pp);
	}while(l->x == Ignore);

	len = p - op;
	Name n = nameofn(op, len);
	if(FLAGS_parsedebug == 1)
		fprint(2, "lex: %d %s\n", tokno+1, s);
	List<GlrToken<AstToken>*> *lt = nil;
	++tokno;
	for(; l; l=l->next){
		sym = l->x;
		t = glrnewtoken(parse, sym);
		a = (AstToken*)t->val;
		a->ast = mkaststring(sym, n);
		a->tokno  = tokno;
		a->tokno1 = 0;
		a->tokno2 = 0;
		a->zline.file = parse->file;
		a->zline.line = parse->line;
		lt = listappend(nil, lt, t);
	}
	return lt;
}

// Parse the cpp's "# ..." directives, assuming that the current
// position is "^".  This will eat all sol whitespace and record
// line number info (if present).  We could do some fancier
// parsing of pragmas, but we won't.
static void
glrlexsharp(GlrParse<AstToken> *parse, char **pp)
{
	char *op, buf[512];
	int line, i;

	assert(pp);
	assert(*pp);
	op = *pp;
	while(**pp && (**pp == ' ' || **pp == '\t'))
		(*pp)++;
	if(**pp != '#')
		goto done;
	(*pp)++;
	// Line number
	while(**pp && (**pp == ' ' || **pp == '\t'))
		(*pp)++;
	line = 0;
	if(**pp < '0' || '9' < **pp)
		goto eatline;
	while(**pp && '0' <= **pp && **pp <= '9'){
		line = line*10 + (**pp - '0');
		(*pp)++;
	}
	parse->line = line - 1;  // the \n will add 1
	// File name
	while(**pp && (**pp == ' ' || **pp == '\t'))
		(*pp)++;
	if(**pp != '"')
		goto eatline;
	(*pp)++;
	for(i=0; **pp && **pp != '"'; i++){
		if(**pp == '\\'){
			(*pp)++;
			if(!**pp)
				break;
		}
		if(i >= (int)sizeof buf){
			buf[sizeof(buf)-1] = '\0';
			panic("file name too long near '%s'", buf);
		}
		buf[i] = **pp;
		(*pp)++;
	}
	buf[i] = '\0';
	parse->file = nameof(buf);

eatline:
	while(**pp && **pp != '\n')
		(*pp)++;
done:
	parse->col += *pp - op;
}

// Builtin function for parsing via GLR.
Ast*
parsebuiltin(Type *type, char *text, Name file)
{
	Ast *ast;
	AstToken *pv;
	GlrParse<AstToken> *p;
	Gram *gram;
	Type *tgram, *tsym, *trule;
	char *t;

	assert(type);
	assert(type->op == TypeSyntax);
	tgram = gramsupertype(type);
	tsym  = symsupertype(type);
	trule = rulesupertype(type);
	assert(tgram);
	assert(tsym);
	gram = tgram->grammar;
	assert(gram);

	p = glrnewparse(&gram->cfg);
	p->file = file;
	p->line = 1;
	p->col  = 1;
	p->merge = glrmerge;
	p->g->debug = (FLAGS_parsedebug) ? FLAGS_parsedebug-1 : 0;
	t = text;
	if((pv = glrparse(&gram->cfg, tsym->cfgsym, p, &glrlexstr, &t)) == nil){
		fprint(2, "%s:%d: error parsing, near %.100s\n",
		          p->file ? (char*)p->file : "???", p->line, t);
		return nil;
	}
	ast = pv->ast;
	// Grammar validation
	if(trule && (ast->rule == nil || ast->rule != trule->cfgrule)){
		fprint(2, "%s:%d: expected rule '%s', parsed as rule '%s'\n",
		          (p->file) ? (char*)p->file : "???", p->line,
		          trule->name, ast->rule ? (char*)ast->rule->name : "<nil>");
		return nil;
	}
	if(!validateast(ast, tgram, 1))
		return nil;
	if(FLAGS_parsedebug)
		fprint(2, "string:\n  %s\n  %T\n %#A\n", text, ast->_specifictype, ast);
//	glrfreeparse(p);
	return ast;
}


// ========================================================
// Parsing patterns

static void patternmerge(CfgSym<AstToken>*, AstToken*, AstToken*);
static List<List<GlrToken<AstToken>*>*>* convertptokens(GlrParse<AstToken>*, Type*, List<PToken*>*, View*);
static List<GlrToken<AstToken>*>* glrlexlist(GlrParse<AstToken>*, CfgGram<AstToken>*, List<List<GlrToken<AstToken>*>*>**);
static Ast* disambiguate(Ast*, Type*);
static Ast* disambiguate1(Ast*, Type*);
static int depth(Ast*);
static void rewritewildcards(Ast*);
static int rewriteptokentypes(Type*, Ast*, List<PToken*> *ptokens, View*);

// Convert a pattern expression to list of list of GlrToken<AstToken> for use by parser.

// Parse the pattern represented by expr.  Set expr->ast to the result.
// Return 0 on failure, 1 on success.  Only update the slot types when
// 'infertypes' is nonzero (e.g., when the pattern has Bind semantics).
int
parsepattern(ExprParse *expr, int infertypes)
{
	AstToken *pv;
	Ast *ast;
	GlrParse<AstToken> *p;
	CfgSym<AstToken> *start;
	Line errline;

	assert(expr);
	assert(expr->type_);
	assert(expr->type_->op == TypePtr);
	assert(expr->type_->left->op == TypeSyntax);
	expr->ast_ = nil;	// defense

	// The assigned type of expr tells us where to start.
	// It can be any of the three Ast subtype levels:
	//   gram: this is an error (we need a start symbol)
	//   sym:  use this as the start symbol
	//   rule: use the super-type (sym) as the start symbol, then
	//         verify that the resulting parse used the same rule.
	Type *tgram = gramsupertype(expr->type_->left);
	Type *tsym  = symsupertype(expr->type_->left);
	Type *trule = rulesupertype(expr->type_->left);
	assert(tgram);

	Gram *gram = tgram->grammar;
	assert(gram);

	if(tsym == nil){
		error(expr->line_, "missing non-terminal start name in %#E", expr);
		return 0;
	}

	View *view = tsym->view;
	if(view)
		start = view->parseme->cfgsym;
	else
		start = tsym->cfgsym;

	// Parse.
	p = glrnewparse(&gram->cfg);
	p->file = expr->line_.file;
	p->line = expr->line_.line;
	p->col  = 1;
	p->merge = patternmerge;
	List<List<GlrToken<AstToken>*>*>* lexed = convertptokens(p, tgram, expr->ptokens_, view);
	if((pv = glrparse(&gram->cfg, start, p, glrlexlist, &lexed)) == nil){
		error(expr->line_, "couldn't parse %#E as %s", expr, start->name);
		memset(&errline, 0, sizeof errline);
		errline.file = p->file;
		errline.line = p->line;
		error(errline, "parse error near this line");
		return 0;
	}

	// Disambiguate wildcards and infer slot types.
	ast = pv->ast;
	ast = disambiguate(ast, tgram);
	rewritewildcards(ast);

	if(isambiguous(ast)){
		error(expr->line_, "couldn't parse %#E as %s: ambiguous %l#A",
		                  expr, start->name, ast);
		return 0;
	}

	// If a rule was given, check that it matches.
	if(trule && !view && (ast->rule == nil || ast->rule != trule->cfgrule)){
		error(expr->line_, "expected rule '%s', parsed as rule '%s'",
		                  trule->name, ast->rule ? (char*)ast->rule->name : "<nil>");
		return 0;
	}

	if(FLAGS_parsedebug)
		fprint(2, "pattern:\n  %#E\n  %T\n %#A\n", expr, expr->type_, ast);

	if(!validateast(ast, tgram, 0))
		return 0;

	if(infertypes && !rewriteptokentypes(tgram, ast, expr->ptokens_, view))
		return 0;

	// Narrow expr->type_ from the sym type to the rule type.
	// Can't use cfgruletype() since that doesn't work with wildcard types.
	if(ast->rule && !view && !trule){
		Type *r = _cfgruletype(ast->rule);
		assert(r);
		Type *t = lookuptag(tsym, r->name);
		assert(t);
		expr->type_ = typeptr(t);
	}

	expr->ast_ = ast;
	return 1;
}

// FIXME: think carefully about the types passed to mkastslot()
//   and mkastlistslot() ... i think these are right as is ...
// Create a GlrToken<AstToken> for a slot of type sym.
static GlrToken<AstToken>*
mkslot(GlrParse<AstToken> *parse, Type *type, Name slotname)
{
	AstToken *at;
	GlrToken<AstToken> *gt;

	assert(type && type->super && !type->super->super);
	assert(type->cfgsym);

	gt = glrnewtoken(parse, type->cfgsym);
	at = (AstToken*)gt->val;
	at->ast = mkastslot(lookuptag(type, nameof("slot")), slotname);
	//at->ast = mkastslot(type, slotname);
	return gt;
}

static GlrToken<AstToken>*
mklistslot(GlrParse<AstToken> *parse, Type *type, Name slotname)
{
	AstToken *at;
	GlrToken<AstToken> *gt;

	assert(type && type->super && !type->super->super);
	assert(type->cfgsym);

	gt = glrnewtoken(parse, type->cfgsym);
	at = (AstToken*)gt->val;
	at->ast = mkastlistslot(lookuptag(type, nameof("slot")), slotname);
	//at->ast = mkastlistslot(type, slotname);
	return gt;
}

// Translate view slots to the appropriate internal slots
// using slotsin[symname]
static List<GlrToken<AstToken>*>*
translateviewslots(GlrParse<AstToken> *parse, Type *tgram, View *view, PToken *ptok, Name slot)
{
	Type *type;
	Name name;

	List<GlrToken<AstToken>*> *lg = ZList;

	// Try to translate the sym first
	// eg, ::Type --> ::fnarg
	if(ptok->symname){
		List<CfgSym<AstToken>*> *l;
		if((l = hashget(view->slotins, ptok->symname)) != nil){
			for(; l; l=l->next){
				type = _cfgsymtype(l->x);
				type = lookuptag(tgram, type->name);
				lg = listappend(nil, lg, mkslot(parse, type, slot));
			}
			return lg;
		}
	}

	// Next try to translate the element sym
	// eg, ::Type[,]* --> ::@fnarg
	if(ptok->elemname){
		List<CfgSym<AstToken>*> *l;
		if((l = hashget(view->slotins, ptok->elemname)) == nil)
			return nil;
		int have=0;
		for(; l; l=l->next){
			type = _cfgsymtype(l->x);
			name = nameprint("@%s", type->name);
			if((type = lookuptag(tgram, name)) == nil)
				continue;
			have = 1;
			lg = listappend(nil, lg, mklistslot(parse, type, slot));
		}
		if(!have){
			error(ptok->expr->line_, "no grammar symbol is a list of %s", view->viewtype->name);
			return nil;
		}
		return lg;
	}

	return nil;
}

// Lex list of PatternTokens into list of GlrTokens.
// Actually, since GLR admits ambiguous lexing, convert into
// list of list of GlrTokens.
// TODO: rsc tried to clean this up, but it's gotten unwieldy again.
// Do something to make the logic clearer.
static List<List<GlrToken<AstToken>*>*>*
convertptokens(GlrParse<AstToken> *parse, Type *tgram, List<PToken*> *ptokens, View *view)
{
	char *s;
	CfgGram<AstToken> *g;
	GlrToken<AstToken> *gt;
	Gram *gram;
	List<GlrToken<AstToken>*> *lg;
	List<List<GlrToken<AstToken>*>*> *ll;
	Name name;
	Name slot;
	PToken *ptok;
	Type *type;

	gram = tgram->grammar;
	g = &gram->cfg;

	ll = ZList;
	for(; ptokens; ptokens=ptokens->next){
		ptok = ptokens->x;

		if(ptok->op == PTokenString){
			// Parse string using glrlexstr, collecting the results.
			s = ptok->str;
			for(;;){
				lg = glrlexstr(parse, g, &s);
				if(lg == nil){
					error(ZL, "error lexing - %s", ptok->str);
					return nil;
				}
				gt = lg->x;
				if(gt->sym == gram->eof && *s == 0)
					break;
				ll = listappend(nil, ll, lg);
			}
			continue;
		}

		// Otherwise is an expression slot, with type-dependent meaning.
		assert(ptok->op == PTokenExpr);

		ExprName *en;
		if((en = ptok->expr->exprname()) != nil)
			slot = en->name_;
		else
			slot = nameprint("(%#E)", ptok->expr);

		// typecheckslottoken() has filled in some helpful fields of ptok.
		// Examples:
		//    t                symname   elemname  islist
		//  ------------------------------------------------
		//  ???                  nil       nil       0
		//  ptr G                nil       nil       0
		//  ptr G.expr          "expr"     nil       0
		//  ptr G.expr?        "expr?"     nil       0
		//  ptr G.expr[,]+    "expr[,]+"   nil       0 
		//  list ptr G           nil       nil       1
		//  list ptr G.expr      nil      "expr"     1

		if(ptok->elemname)
			assert(ptok->islist);

		// If a view is active, translate slots which have type <: viewtype
		// or type <: reptype to the appropriate internal slots.
		if(view){
			if((lg = translateviewslots(parse, tgram, view, ptok, slot)) != nil){
				ll = listappend(nil, ll, lg);
				continue;
			}
		}

		// Go ahead and use the symname, if known.
		if(ptok->symname){
			if((type = lookuptag(tgram, ptok->symname)) == nil){
				error(ptok->expr->line_, "unknown grammar symbol %s", ptok->symname);
				return nil;
			}
			if(ptok->islist)
				ll = listappend(nil, ll, list1(nil, mklistslot(parse, type, slot)));
			else
				ll = listappend(nil, ll, list1(nil, mkslot(parse, type, slot)));
			continue;
			continue;
		}

		// If a list element name was given, shift @elemname.
		if(ptok->elemname){
			name = nameprint("@%s", ptok->elemname);
			if((type = lookuptag(tgram, name)) == nil){
				error(ptok->expr->line_, "no grammar symbol is a list of %s", ptok->elemname);
				return nil;
			}
			ll = listappend(nil, ll, list1(nil, mklistslot(parse, type, slot)));
			continue;
		}

		// Otherwise, shift a wildcard.
		if(ptok->islist){
			type = lookuptag(tgram, gram->anylist->name);
			ll = listappend(nil, ll, list1(nil, mklistslot(parse, type, slot)));
		}
		else{
			type = lookuptag(tgram, gram->any->name);
			ll = listappend(nil, ll, list1(nil, mkslot(parse, type, slot)));
		}
	}
	if(FLAGS_parsedebug > 2){
		for(List<List<GlrToken<AstToken>*>*> *a=ll; a; a=a->next)
			for(List<GlrToken<AstToken>*> *b=a->x; b; b=b->next){
				gt = b->x;
				fprint(2, "lextok: %A\n", ((AstToken*)gt->val)->ast);
			}
	}
	return ll;
}

// Return next token in the input.
// The state, in vstate, points at a List of List of GlrToken
// that we just pick off one at a time.
static List<GlrToken<AstToken>*>*
glrlexlist(GlrParse<AstToken> *parse, CfgGram<AstToken> *g, List<List<GlrToken<AstToken>*>*> **lp)
{
	static int tokno;

	List<List<GlrToken<AstToken>*>*> *l = *lp;
	if(l == nil)
		return list1(nil, glrnewtoken(parse, ((Gram*)g)->eof));
	
	List<GlrToken<AstToken>*> *out = l->x;	// list of GlrToken
	*lp = l->next;
	free(l);

	++tokno;
	for(List<GlrToken<AstToken>*>* ll = out; ll; ll=ll->next)
		((AstToken*)ll->x->val)->tokno = tokno;

if(0){
	print("lex:");
	GlrToken<AstToken> *t;
	AstToken *a;
	for(List<GlrToken<AstToken>*> *l=out; l; l=l->next){
		t = l->x;
		a = (AstToken*)t->val;
		print(" %#A", a->ast);
	}
	print("\n");
}

	return out;
}

// GLR merge function for pattern parsing.
static void
patternmerge(CfgSym<AstToken> *sym, AstToken *a, AstToken *b)
{
	// We could use a more aggressive merge function here
	// rather than save the trimming for after the parse,
	// but the original implementation in tinyrscc does it this way.
	// (It has code to do it the aggressive way, but that is 
	// commented out, and I have no idea why.  It is certainly
	// worth trying to figure this out at some point in the future.
	// It might have to do with the first pass of disambiguate 
	// below.)
	glrmerge(sym, a, b);
}

// Is this a node of the form (foo (any)) ?
// If so, return the (any).
static Ast*
iswildcard(Ast *a)
{
	Ast *r;

	if(a == nil || a->nright != 1)
		return nil;
	r = astright(a, 0);
	if(r->sym->name[0] == '.' || r->sym->name[0] == '@')
		return r;
	return nil;
}

// The pattern will likely have many ambiguities involving wildcards,
// for example both (expr (wildCARD)) and (expr (number (wildCARD)).
// Disambiguate by choosing shallowest parse.
// TODO: Note that this unfortunately disambiguates parses
// having nothing to do with wildCARDs.
static Ast*
disambiguate(Ast *a, Type *tgram)
{
	Ast *b;

	// First step: if we were parsing something like
	// expr{\a}, then we'll get (expr (wildCARD)) as our simplest
	// parse, which is degenerate.  We want something 
	// a little more interesting.
	while(a->op == AstMerge){
		if(iswildcard(astright(a, 0)))
			b = astright(a, 1);
		else if(iswildcard(astright(a, 1)))
			b = astright(a, 0);
		else
			break;
		a = b;
	}
	return disambiguate1(a, tgram);
}

// Remove merge nodes by eliminating type constraints
// and choosing lowest depth.
static Ast*
disambiguate1(Ast *ast, Type *tgram)
{
	Ast *a, *b;
	int i, d0, d1;
	
	for(i=0; i<ast->nright; i++)
		ast->_right[i] = disambiguate1(astright(ast, i), tgram);

	if(ast->op != AstMerge)
		return ast;
	a = astright(ast, 0);
	b = astright(ast, 1);

	// Resolve by type.
	if(tgram){
		d0 = validateastrule(a->rule, tgram);
		d1 = validateastrule(b->rule, tgram);
		if(d0 && !d1)
			return a;
		if(d1 && !d0)
			return b;
	}

	// Resolve by depth.
	d0 = depth(a);
	d1 = depth(b);
	if(d0 < d1)
		return a;
	else if(d1 < d0)
		return b;
	return ast;
}

// Compute depth of ast tree for use in unmerging.
// Leaves have depth 1.
static int
depth(Ast *a)
{
	int i, n, m;

	if(a == nil)
		return 0;
	switch(a->op){
	default:
		panic("depth");

	case AstString:
		return 1;

	case AstMerge:
		// Pretend merge is not there.
		// (This case doesn't happen when being
		// called by disambiguate1 anyway.)
		n = depth(astright(a, 0));
		m = depth(astright(a, 1));
		if(n < m)
			return n;
		return m;

	case AstRule:
		// 1 + max(kids)
		n = 0;
		for(i=0; i<a->nright; i++){
			m = depth(astright(a, i));
			if(m > n)
				n = m;
		}
		return n+1;

	case AstSlot:
	case AstListSlot:
		return 1;
	}
}

// Rewrite (expr slot) into a slot of type expr.
// Also do this for typeslots.
static void
rewritewildcards(Ast *a)
{
	int i;
	Ast *r;

	if(a == nil || !a->hasslots)
		return;
	// Special case: list slots end up like (expr* (expr@ (@expr))).
	// Rewrite this to (expr*.slot).
	if(astlistkind(a->sym) && a->nright == 1 && (r = iswildcard(astright(a, 0))) != nil){
		a->op = r->op;
		a->text = r->text;
		a->_specifictype = lookuptag(symsupertype(a->_specifictype), nameof("slot"));
		return;
	}
	// Normal (expr slot) case
	if((r = iswildcard(a)) != nil){
		a->op = r->op;
		a->text = r->text;
		a->_specifictype = lookuptag(symsupertype(a->_specifictype), nameof("slot"));
		return;
	}
	for(i=0; i<a->nright; i++)
		rewritewildcards(astright(a, i));
}

// Are there any ambiguous (merge) nodes left?
int
isambiguous(Ast *a)
{
	int i;

	if(a == nil)
		return 0;
	if(a->op == AstMerge)
		return 1;
	for(i=0; i<a->nright; i++)
		if(isambiguous(astright(a, i)))
			return 1;
	return 0;
}

// Now that the slots have been decided unambiguously,
// set the inferred type of each ptoken.
static int _rewriteptokentypes(Type *tgram, Ast *a, List<PToken*> **ptokens, View *view);

static int
rewriteptokentypes(Type *tgram, Ast *a, List<PToken*> *ptokens, View *view)
{
	return _rewriteptokentypes(tgram, a, &ptokens, view);
}

static int
_rewriteptokentypes(Type *tgram, Ast *a, List<PToken*> **ptokens, View *view)
{
	int i;
	PToken *pt;
	Name name;

	if(a == nil)
		return 1;

	if(a->op == AstSlot){
		// Find next PTokenExpr in list.
		for(;;){
			if(*ptokens == nil)
				panic("_rewriteptokentypes");
			pt = (*ptokens)->x;
			*ptokens = (*ptokens)->next;
			if(pt->op == PTokenExpr)
				break;
		}

		// If pt->expr is a typeless name, assign the correct type.
		ExprName *en;
		if((en = pt->expr->exprname()) == nil || en->type_ != nil)
			return 1;

		Type *t;
		// First check if the type should be translated by a view.
		if(view && (name = hashget(view->slotouts, a->symname)) != nil){
			t = lookuptag(tgram, name);
			assert(t);
		}
		// Otherwise just use the symtype.  Make sure it
		// is a direct subtype of the pattern's grammar
		// (can't use cfgsymtype() because a->sym might not belong to tgram)
		else{
			t = lookuptag(tgram, a->symname);
			assert(t);
		}

		pt->expr->type_ = typeptr(t);
		if(FLAGS_parsedebug)
			fprint(2, "  %#E :: %T\n", pt->expr, pt->expr->type_);
		return 1;
	}

	// Should never be a list slot.
	assert(a->op != AstListSlot);

	for(i=0; i<a->nright; i++)
		if(!_rewriteptokentypes(tgram, astright(a, i), ptokens, view))
			return 0;
	return 1;
}


// ========================================================
// Grammar rule validation
//
// Assert that all rules in a syntax tree originate from
// grammars that are a subtype of the grammar type 'tgram'.
// If 'allowwildcard' is false, we'll remove the wildcard
// from 'tgram' if it exists.  This is used by parse patterns.

static int
isnotwildcardtype(Type *t)
{
	return t == nil || t->op != TypeDontCare;
}

static int _validateast(Ast*, Type*);

static int
validateast(Ast *ast, Type *tgram, int allowwildcard)
{
	Type *t;
	int iswildcard = 0;

	// Look for a wildcard
	if(tgram->op == TypeUnion){
		for(List<Type*> *l=tgram->types; l; l=l->next){
			t = l->x;
			if(t->op == TypeDontCare){
				iswildcard = 1;
				break;
			}
		}
	}

	// If wildcards are allowed and we have one, everything is ok.
	if(iswildcard && allowwildcard)
		return 1;

	// If we have a wildcard, remove it.
	// Fast case: (G+?) --> G
	if(tgram == tgram->grammar->wildroottype)
		tgram = tgram->grammar->roottype;
	// General case
	else if(iswildcard){
		List<Type*> *l = listcopy(nil, tgram->types);
		List<Type*> *l1 = listfilter(l, isnotwildcardtype);
		tgram = typeunion(l1, tgram->grammar);
		listfree(l);
		listfree(l1);
	}

	return _validateast(ast, tgram);
}

static int
_validateast(Ast *ast, Type *type)
{
	int i;

	if(ast == nil)
		return 1;

	switch(ast->op){
	case AstRule:
		if(!validateastrule(ast->rule, type))
			goto Bad;
		// fall through
	case AstMerge:
		for(i=0; i < ast->nright; i++)
			if(!_validateast(astright(ast, i), type))
				return 0;
		return 1;

	default:
		return 1;
	}

Bad: {
	Line line;
	
	line.file = ast->zline.file;
	line.line = ast->zline.line;
	error(line, "parsing constraint violated for type %T", type);
	error(line, "found rule '%T'", _cfgruletype(ast->rule));
	}
	return 0;
}

static int
validateastrule(CfgRule<AstToken> *rule, Type *tgram)
{
	Type *r, *t;

	if(!rule)
		return 0;
	// The origin type of this rule must be a subtype of 'tgram'.
	// Assume that 'rule' and 'tgram' belong to the same family.
	// This assumption will break if we ever allow embedding one
	// grammar in another (a la Jeanie).
	r = _cfgruletype(rule);
	assert(r);
	assert(r->super);
	assert(r->super->super);
	t = r->super->super;
	// We know both 'tgram' and 't' are grammar types.
	// If 'tgram' is not a union, we can use a faster path.
	if(tgram->op == TypeSyntax)
		return issubtypegrammar(tgram, t);
	// Perf note: this could be optimized; most of issubtypesyntax()
	// is useless since we know that both 'tgram' and 't' are grammar
	// types.  What we want starts at the line "// Grammar subtyping.".
	return issubtypesyntax(tgram, t);
}

