// Types for grammars, symbols, and rules
//
// ---------- Grammar Familes -----------
//
// All grammars in zeta are divided into families.  A grammar
// family is backed by a single CFG, so all grammers in the
// same family share the same namespace of symbols, rules, and
// attributes.
//
// Each family has a "root" grammar.  All other grammars in a
// family are modular extensions of that root grammar.  For
// example, the following shows a single grammar family rooted
// at "C":
//
//    grammar C {
//      expr: <Plus> expr "+" expr
//      ...
//    }
//
//    grammar Rotate extends C {
//      expr: expr "<<<" expr
//      ...
//    }
//
//    grammar Lambda extends C {
//      expr: "fn" fndef;
//    }
//
// (Note that all root grammars are "extensible" so the keyword
// is optional, a la zeta structs.)
//
// Each grammar (in this case, "C", "Rotate", and "Lambda") adds
// a new name into the current scope with kind TypeSyntax.  There
// are also subtypes for symbols (C.expr) and rules (C.expr.Plus).
// These are discussed below.
//
// The special root grammar "Ast" is not a grammar at all.  it
// is an abstract type which cannot be parsed, and so it cannot be
// used in any syntax patterns.  Note that attributes are attached
// to a grammar family; the only attributes accessible from the
// abstract type "Ast" are the builtin attributes "prev" and "next".
// Further, the only direct subtypes of "Ast" are the special types
// "Ast.slot" and "Ast.merge", which denoted Asts of type AstSlot
// and AstMerge respectively.
//
// ------------- Union Types ------------
//
// Grammars in the same family can be unioned together, for example:
//
//    C                       // plain old C
//    (C + Rotate)            // ... plus Rotate
//    (C + Rotate + Labmda)   // ... plus Lambda
//
// A union type, represented by TypeUnion, is the set-theoretic
// union of all rules in all unioned grammars.
//
// The special wildcard "?" can be used in a union type as a sort of
// existential.  It marks the presence of an unknown grammar in the
// union.  For example:
//
//    (C + ?)                // plain old C plus unknown stuff
//
// Additionally, it is possible to import multiple extension grammars
// when extending a root grammar.  For example:
//
//    grammar Silly extends C, Rotate, Lambda { ... }
//
// (Note that all grammars must belong to the same family.)
// This is equivalent to:
//
//    grammar Silly2 extends C { ...  }
//    typedef Silly: (C + Silly2 + Rotate + Lambda);
//
// As it turns out, zeta currently has no direct syntax for union
// types.  In fact, only two kinds of types can be expressed with
// the current syntax:
//
//     C      // plain old grammar C
//     C?     // same as (C + ?) about
//
// Union types must be named via a grammar statement, like in
// the following:
//
//    grammar CRotateLambda extends C, Rotate, Lambda {}
//
// ------------- Subtyping --------------
//
// There are three kinds of subtyping.
//
// 1.Subtyping amongst grammars, symbols and rules:
//      G.sym.rule <: G.sym <: G
//
// 2.Subtyping amongst grammars in the same family.  The root grammar
//   of a family is the natural "bottom" of this relation.  The rest
//   of the relation is formed by a DAG of "extends" relationships.
//   Union types (and multiply-extended grammars like Silly) also
//   exist in this relation.  For example:
//                  C             // root grammar is the Bottom
//                /   \           //
//           Rotate   Lambda      // A <: B iff exists. path from A down to B
//                \  /
//                 \/
//           (Rotate + Lambda)
//                 |
//             (... + ?)          // anything with ? is at the Top
//
//   This subtyping further extends to symbols and rules:
//
//               A <: B
//           --------------
//           A.sym <: B.sym
//
// 3.Finally, the abstract type "Ast" is treated specially; all grammars
//   are a subtype of "Ast":
//      forall G. G <: Ast
//
// ----------- Representation -----------
//
// There are three parallel heirarchies, types, structreps, and cfgs:
//
//                        Ast Struct
//                          |
//    Grammar Type    Grammar Struct    Cfg
//          |               |             |
//     Symbol Type     Symbol Struct    CfgSym
//          |               |             |
//       Rule Type       Rule Struct    CfgRule
//
// There is a single "struct" and "cfg" heirarchy per grammar family,
// and there is one type heirarchy per grammar type.
//
// "Structs" store annotations on syntax objects.  These include
// (immutable) attributes and (mutable) fields.  Since structs are shared
// within an entire family, so are annotations.  They can be localized to
// a specific rule or symbol, and subtyping applies: all annotations at
// the grammar level are inherited at the symbol level, and similarly for
// the symbol and rule levels.  For example:
//
//    attribute G          :: foo;  // shared by G::foo, G.sym::foo, G.sym.rule::foo
//    attribute G.sym      :: bar;  // shared by G.sym::bar, G.sym.rule::bar
//    attribute G.sym.rule :: baz;  // local to G.sym.rule::baz
//
// Note from the diagram above that there is a global "Ast" struct.
// That struct stores annotations which are shared among all grammar
// families, like the builtin "prev" and "next" attributes.  It is
// possible to extend the "Ast" struct with new fields, but because
// of the way data is layed out (see xattr), all global attributes
// must be builtin; you cannot add more global attributes.
//
// Each grammar type (a TypeSyntax or TypeUnion) points to its
// family's struct object using the Type::structrep field.
//
// Also, the fields Type::grammar, Type::cfgsym, and Type::cfgrule
// are used to link to the underlying Gram*, CfgSym*, and CfgRule*
// for grammar, symbol, and rule types.
//
// ------------ Construction ------------
//
// We take the following steps to construct a grammar exension X:
//
// * Make a new TypeSyntax named X, added to the local scope.
//
// * For each extended grammar G, copy all G.sym and G.sym.rule into
//   new types X.sym and X.sym.rule.  (This only copies the types, not
//   the underlying CFG; the new types point to the same syms and
//   rules.)
//   * Panic on conflict, such as a symbol named "foo" which is uniquely
//     defined in grammar extensions A and B.  Currently there is no way
//     to resolve this confict; a simple method would be to explicitly
//     prefix the symbol, as in A.foo or B.foo.
//
// * For each new symbol "sym" in X, add a CfgSym named "X.sym" into
//   the family's CFG.  Similarly, each new rule is named "X.rule".
//   This way, symbol and rule names are locally scoped in the CFG.
//
// -------------- Scoping ---------------
//
// TODO: anything interesting to say about how grammar names are scoped?
// TODO: new symbol and rule names are scoped locally,  eg,
//           grammar Foo extends C {
//              newsym: ...;        // the name C.newsym is hidden as Foo.newsym
//           }
//       (this is sort of said above)
//
// FIXME: XXX XXX XXX
//  Scoping doesn't really work at all because typesubname() is
//  fundamentally broken -- it needs to know about scopes.  Fixing
//  this would require a significant rewrite of the parser.  We'd
//  have to create scopes *at parse time* so that typesubname()
//  can know which scope is currently active.  Note that typename()
//  is not affected because it does not have to be canonicalized
//  in the same tricky way as typesubname() -- see mksyntaxsubtype().
//
// --------------- Views ----------------
//
// Views are ugly and hopefully they are replaced by something less
// ugly.  Therefore I won't try to describe them here.
//

#include "grammar.h"

// see comment in grammarbuild.c
#define lookuptag(a, b)   lookuptag(a, b, nil)

Type*
gramroottype(struct Gram *g)
{
	return g ? g->roottype : nil;
}

Type*
gramwildtype(struct Gram *g)
{
	return g ? g->wildroottype : nil;
}

static void
assertgramtype(Type *gramtype)
{
	assert(gramtype);
	assert(gramtype->op == TypeSyntax || gramtype->op == TypeUnion);
	assert(gramtype->super == nil);
}

static void
assertsymtype(Type *symtype)
{
	assert(symtype);
	assert(symtype->op == TypeSyntax);
	assert(symtype->super != nil);
	assert(symtype->super->super == nil);
}

// ========================================================
// Type construction: Brand New Types

static Type*
mksyntaxstructrep(Name name, Type *superrep)
{
	// Construct a TypeStruct representation for a syntax
	// type.  Root structs inherit from Ast->structrep;
	// everyone else inherits from 'superrep'.
	Type *structrep;

	if(superrep == nil){
		Type *t = typeast()->structrep;
		structrep = mkTypeStruct(nil, name, t);
		structrep->attrsize = t->attrsize;  // t->attrsize is fixed!
	}else{
		structrep = mkTypeStruct(nil, name, superrep);
		structrep->declscope = superrep->declscope;
		addstructtag(superrep, structrep);
	}
	structrep->tagshash = mkhash<Name,List<Type*>*>();
	structrep->extensible = 1;
	structrep->discriminated = 1;
	return structrep;
}

static Type*
mksyntaxgramtype(Name name, Type *structrep, Gram *g)
{
	assert(name);
	assert(g);

	// Construct a brand-spanking new grammar type which
	// belongs to grammar family 'g'.  If 'structrep' is
	// non-nil, this is a grammar extension.
	if(!structrep)
		structrep = mksyntaxstructrep(name, nil);
	Type *t = mkTypeSyntax(name, nil, structrep, g);
	t->tagshash = mkhash<Name,List<Type*>*>();
	return t;
}

static void
fillinsyntaxsubtype(Type *t, Type *super, Type *structrep)
{
	assert(t);
	assert(t->op == TypeSubname);
	assert(super);
	assert(structrep);
	assert(structrep->op == TypeStruct);

	t->op = TypeSyntax;
	t->structrep = structrep;
	t->grammar   = super->grammar;
	t->tagshash  = mkhash<Name,List<Type*>*>();
	t->declscope = super->declscope;
	addstructtag(super, t);
}

static Type*
mksyntaxsubtype(Name name, Type *super, Type *structrep)
{
	assert(super);

	// Construct a brand-spanking new symbol or
	// rule type which belongs to grammar family
	// 'g' and is a direct subtype of 'super'.
	if(!structrep)
		structrep = mksyntaxstructrep(name, super->structrep);

	// User code parses types like 'Gram.expr' as a TypeSubname.
	// These are resolved lazily in-place by typecheck.c to
	// TypeSyntaxs.  To maintain Type canonicalization, we need
	// to emulate this behavior here: first build a TypeSubname,
	// then "resolve" it to a TypeSyntax.  Later calls to
	// typesubname() will return this type, even though we've
	// changed its 'op'.
	Type *t = typesubname(super, name);

	// Shouldn't be resolved yet!
	assert(t->op == TypeSubname);

	// Now resolve it!
	fillinsyntaxsubtype(t, super, structrep);
	return t;
}

#define mkTypeSyntax DONT_CALL_mkTypeSyntax

Type*
addgramtype(Gram *g, Name name, Scope *scope)
{
	Type *type;

	// New grammar family?
	if(!g->roottype){
		type = mksyntaxgramtype(name, nil, g);
		g->roottype = type;
	}else{
		type = mksyntaxgramtype(name, g->roottype->structrep, g);
	}

	type->declscope = scope;
	type->structrep->declscope = scope;
	return type;
}

Type*
addsymtype(Type *gramtype, CfgSym<AstToken> *s, Name name)
{
	Type *type;

	assertgramtype(gramtype);

	type = mksyntaxsubtype(name, gramtype, nil);
	type->cfgsym = s;
	s->_origin = type;
	addbuiltingramtags(type, s, s);
	return type;
}

Type*
addruletype(Type *symtype, CfgRule<AstToken> *r, Name name)
{
	Type *type;

	assertsymtype(symtype);

	type = mksyntaxsubtype(name, symtype, nil);
	type->cfgrule = r;
	r->_origin = type;
	addbuiltingramtags(type, r, r);
	// If the symbol is mirrored, also add this rule as a tag
	// of the mirror (this is not actually a "view" - views
	// exist at the symbol level - so we leave that field empty).
	if(symtype->mirroredby){
		Type *m;
		m = addruletype(symtype->mirroredby, r, name);
		m->mirrors = type;
		type->mirroredby = m;
	}
	return type;
}

void
addbuiltingramtags(Type *type, void *cfgslot, void *cfgmerge)
{
	assert(type->op == TypeSyntax);
	assert(type->structrep);

	// TODO: also make custom struct reps?
	Type *tslot  = mksyntaxsubtype(Name::slot,  type, type->structrep);
	Type *tmerge = mksyntaxsubtype(Name::merge, type, type->structrep);
	if(cfgslot){  // XXX
		tslot->cfgsym  = (CfgSym<AstToken>*)cfgslot;
		tslot->cfgrule = (CfgRule<AstToken>*)cfgslot;
	}
	if(cfgmerge){
		tmerge->cfgsym  = (CfgSym<AstToken>*)cfgmerge;
		tmerge->cfgrule = (CfgRule<AstToken>*)cfgmerge;
	}
}

// ========================================================
// Type construction: Lazy Subtype Resolution

static View* _mkgramview(Type*, Type*, CfgSym<AstToken>*);

static Type*
findsyntaxsubtype(Name name, List<Type*> *types, List<Type*> **all)
{
	Type *rep = nil;
	Type *old = nil;
	*all = nil;

	for(List<Type*> *l=types; l; l=l->next){
		Type *super = l->x;
		if(super->op == TypeDontCare)
			continue;
		List<Type*>* ll = hashget(super->tagshash, name);
		// If there is a tag, there should only be one!
		Type *cur;
		if(ll){
			cur = ll->x;
			assert(ll->next == nil);
		}
		// Not found?  Try to resolve it.  (This might be expensive!)
		else{
			cur = typesubname(l->x, name);
			if(cur->op == TypeSubname)
				continue;
		}
		// Check for a conflict.
		assert(cur);
		if(!rep)
			rep = cur->structrep;
		else if(cur->structrep != rep)
			panic("conflicting definition of name: %s (also in %s)", name, super);
		*all = listappend(nil, *all, cur);
		old = cur;
	}

	return old;
}

void
resolvesyntaxsubtype(Name name, Type *super, Type *canon)
{
	assert(super);
	assert(canon);

	// Given a TypeSyntax or TypeUnion type 'super', lazily
	// resolve its subtype 'name'.  The canonical object for
	// this subtype is 'canon', so we modify that directly.

	// Search for a resolution.  The meaning of 'super->types'
	// depends on 'super->op':
	//  TypeSyntax - the list of types that 'super' extends
	//  TypeUnion  - the list of types in the union
	if(!super->types)
		return;

	List<Type*> *deps = nil;
	Type *old;
	Gram *g = super->grammar;

	// Special case: (G+?) types.
	// This is essentially a direct copy of G.
	if(g && super == g->wildroottype){
		List<Type*>* ll = hashget(g->roottype->tagshash, name);
		// If there is a tag, there should only be one!
		if(ll){
			assert(ll->next == nil);
			old = ll->x;
			deps = list1(nil, old);
		}
		// Not found?  It doesn't exist.
		else
			return;
	}
	// General case.
	else{
		old = findsyntaxsubtype(name, super->types, &deps);
		if(!old)
			return;
	}
	canon->types = deps;

	// Copy 'old' into 'canon'.
	// Views (yuck ...)
	if(old->view){
		Type *rep;
		rep  = typesubname(super, old->view->reptype->name);
		assert(rep->op == TypeSyntax);

		Type *parseme;
		parseme  = typesubname(super, old->view->parseme->name);
		assert(parseme->op == TypeSyntax);

		fillinsyntaxsubtype(canon, super, old->structrep);

		View *view;
		view = _mkgramview(canon, rep, old->cfgsym);
		view->parseme   = parseme;
		view->slotins   = old->view->slotins;   // these don't change so just copy them
		view->slotouts  = old->view->slotouts;
		view->xformname = old->view->xformname;
		view->xform     = old->view->xform;
		view->postxformname = old->view->postxformname;
		view->postxform     = old->view->postxform;
		assert(view->parseme);
		assert(canon->mirrors);
	}
	// Normal symbols
	else{
		fillinsyntaxsubtype(canon, super, old->structrep);
		canon->cfgsym  = old->cfgsym;   // copy both to emphasize that we're
		canon->cfgrule = old->cfgrule;  // copying the union, not an individual item
		// More views (yuck ...)
		if(super->mirrors){
			canon->mirrors = typesubname(super->mirrors, name);
			canon->mirrors->mirroredby = canon;
		}
	}
}

// ========================================================
// Union types

void
adduniongrammar(Type *gramtype, Gram *g)
{
	assert(gramtype);
	assert(gramtype->op == TypeUnion);

	// This is a raw type created by mkTypeUnion(),
	// so we need to set some basics.
	gramtype->structrep = g->roottype->structrep;
	gramtype->tagshash = mkhash<Name,List<Type*>*>();

	// Nothing else to do here; subtypes are created
	// lazily.
}

// ========================================================
// View types (yuck...)

static View*
_mkgramview(Type *viewtype, Type *reptype, CfgSym<AstToken> *sym)
{
	View *view;
	Gram *g;

	assertsymtype(viewtype);
	assertsymtype(reptype);
	g = viewtype->grammar;

	view = (View*)emalloc(sizeof *view);
	view->viewtype = viewtype;
	view->reptype  = reptype;
	view->viewtype->view = view;
	view->viewtype->mirrors   = view->reptype;
	view->reptype->mirroredby = view->viewtype;
	view->slotins  = mkhash<Name,List<CfgSym<AstToken>*>*>();
	view->slotouts = mkhash<Name,Name>();
	view->viewtype->cfgsym = sym;
	return view;
}

View*
mkgramview(Type *gramtype, Name name, Type *reptype)
{
	Gram *g;
	CfgSym<AstToken> *sym;
	Type *type;
	View *view;

	assertgramtype(gramtype);
	assertsymtype(reptype);

	g = gramtype->grammar;
	sym  = cfgnewsym(&g->cfg, nameprint("%s.%s#view", gramtype->name, name), nil);
	type = mksyntaxsubtype(name, gramtype, reptype->structrep);
	view = _mkgramview(type, reptype, sym);
	type->types = list1(nil, view->reptype);
	return view;
}

// ========================================================
// Scoping

static List<Type*>*
gramsinscope(Scope *scope, Gram *g, List<Type*> *grams)
{
	// Gather all in-scope grammars in family 'g'.
	if(!scope)
		return grams;

	for(List<Type*> *l=scope->grams; l; l=l->next){
		Type *t = l->x;
		assert(t->op == TypeSyntax);
		if(t->grammar == g)
			grams = listappend(nil, grams, t);
	}

	grams = gramsinscope(scope->outer, g, grams);
	for(List<Scope*> *l=scope->imported; l; l=l->next)
		grams = gramsinscope(l->x, g, grams);

	return grams;
}

Type*
resolvedefaultxgram(Typecheck *tc, Type *type)
{
	List<Type*> *grams;

	// Resolve (G+?) types to (G+A+..+Z+?), where A..Z
	// are all the grammar extensions in the same family
	// as G that are in scope.

	// We expect that "type" is a grammar type.
	assert(type);
	assert(type->super == nil);

	// Don't resolve abstract grammars (ie, "Ast").
	if(!type->grammar)
		return type;
	// Only resolve (G+?).  Really, we should resolve
	// (G+A+..+?), but user code has no way of writing
	// any types except (G+?).
	if(type != type->grammar->wildroottype)
		return type;

	// Add all grammars in scope.
	grams = listcopy(nil, type->types);
	grams = gramsinscope(tc->scope_, type->grammar, grams);

	Type *u = typeunion(grams, type->grammar);
	listfree(grams);
	return u;
}

Type*
resolvegramwildcard(Line line, Name name)
{
	Type *type;

	// Resolve the type "name + ?"
	type = lookuptype(globalscope, name, 0);
	if(!type){
		warn(line, "(probably spurious): unknown grammar '%s'", name);
		return nil;
	}
	if(type->op != TypeSyntax || type->super || !type->grammar
	|| type->name != type->grammar->name){
		error(line, "expected a root grammar type for '%s', found %T", name, type);
		return nil;
	}

	return type->grammar->wildroottype;
}

