#include <sys/time.h>
#include <sys/resource.h>
#ifdef __linux__
#include <malloc.h>
#endif
#include "a.h"
#include "arg.h"
#include "cfg.h"
#include "slr.h"

DEFINE_int32(maxmem, 500, "Maximum memory footprint (MB)");
DEFINE_bool(memstats, false, "Print memory stats at exit.");
DEFINE_bool(printstmts, false, "Print top-level statements.");
DECLARE_bool(profile);
DECLARE_bool(disassemble);

Arena *asyntax;

Frame *globalframe;
Scope *globalscope;
Scope *realglobalscope;
List<PatternXForm*> *queuedpatternxforms;

static void doexit(void);
static void dopatternxforms(Thread*);
static void loadqueuedextensions(void);
static List<Name> *queuedextensions;
static char *basedir;

void
dostats(void)
{
	system("ps au|grep zeta");
}

int
main(int argc, char **argv)
{
	struct rlimit rlim;
	int done;

	Name::init();
	checkdisasm();
	gflags::SetUsageMessage("usage: zeta filename.zeta [args]");
	gflags::ParseCommandLineFlags(&argc, &argv, true, false);
	argc--; argv++;

	// Limit memory to avoid making
	// entire system swap if we screw up.
	rlim.rlim_cur = FLAGS_maxmem*1000000;
	rlim.rlim_max = FLAGS_maxmem*1000000;
	setrlimit(RLIMIT_AS, &rlim);

	if(argc < 1){
		fprint(2, "usage: zeta [-help] filename.zeta [args]\n");
		exit(1);
	}

	fmtinstall('A', astfmt);
	fmtinstall('E', exprfmt);
	fmtinstall('P', patfmt);
	fmtinstall('R', cfgrulefmt);
	fmtinstall('G', grammarfmt);
	fmtinstall('@', slrstatefmt);
	fmtinstall('$', stmtfmt);
	fmtinstall('T', typefmt);
	fmtinstall('&', locfmt);

	debuginit();

	basedir = estrdup(argv[0]);
	if (strchr(basedir, '/'))
		*(strrchr(basedir, '/')+1) = 0;
	else
		basedir = "./";

	if (strcmp(argv[0], "-") == 0)
		lexinit("/dev/stdin");
	else
		lexinit(argv[0]);
	globalframe = mkFrame();
	globalscope = mkScope(globalframe, nil, nil);
	realglobalscope = globalscope;
	builtininit(argc, argv);

	pusharena(&asyntax);
	do {
		int res;
		showprompt();
		done = 1;
		if((res = yyparse()) < 0)
			panic("yyparse %s", argv[0]);
		if (res > 0 && isinteractive())
			// Parse error.  Suck it up
			done = 0;
	} while (!done);

//	atexit(dostats);
	doexit();

	return 0;
}

extern int nmkast;
void
doexit()
{
	if(FLAGS_memstats){
#ifdef __linux__
		struct mallinfo info = mallinfo();
		fprint(2, "memory: %d\n", info.arena);
#endif
		fprint(2, "gc'ed alloc: %d\n", memtot);
		fprint(2, "frame alloc: %d\n", frametot);
	}

	if(FLAGS_profile)
		printprofile();

	if(debugfd)
		debugbreak();
}

char*
globaldata(void)
{
	static int fsz;
	static char *fp;
	
	if(globalframe->size > fsz){
		fp = (char*)erealloc(fp, globalframe->size);
		memset(fp+fsz, 0, globalframe->size - fsz);
		fsz = globalframe->size;
	}
	return fp;
}

void
showprompt(void)
{
	if (isinteractive()) {
		print("/**/ ");
	}
}

void
dotop(Stmt *stmt)
{
	Thread t;
	static int stickyerrors;

	errors = 0;
	if(FLAGS_printstmts)
		fprint(2, ">> %$\n", stmt);
	Typecheck tc(globalframe, globalscope);
	tc.reachable_ = 1;
	typecheckstmt(&tc, stmt);
	if(errors || stickyerrors){
		if(!isinteractive())
			stickyerrors++;
		return;
	}
	Compile cs(&tc);
	cs.codebuf_ = mkcodebuf();
	compilestmt(&cs, stmt);
	cs.emitC(stmt->line_, xstop);
	patch(&cs);
	aclear(&cs.arena_);
	afree(poparena(&asyntax));
	pusharena(&asyntax);

	if(FLAGS_disassemble)
		disasm(cs.codebuf_->bp, cs.codebuf_->p);

	memset(&t, 0, sizeof t);
	t.depth_ = 10000;		// counts down to zero then boom!
	t.gp_ = globaldata();
	t.fp_ = nil;

	dopatternxforms(&t);
	if(errors || stickyerrors){
		stickyerrors++;
		aclear(&cs.arena_);
		return;
	}

	t.pc_.c = (Code**)cs.codebuf_->bp;
	run(&t);
	loadqueuedextensions();
}

void
doinclude(char *file)
{
	if(pushinput(file, nil)){
		pusharena(&asyntax);
		if(yyparse() < 0)
			panic("yyparse %s", file);
		afree(poparena(&asyntax));
		popinput();
	}
}

void
dorequire(Line line, Name file, int linkin)
{
	static Hash<Name,Scope*> *opened;
	static List<Name> *opening;
	Scope *scope, *old;

	if(!opened)
		opened = mkhash<Name,Scope*>();

	if(file[0] == '.' && file[1] == '/')
		file = nameprint("%s.zeta", file);
	else
		file = nameprint("x/%s.zeta", file);

	// If it already exists, just link it in.
	if((scope = hashget(opened, file)) != nil){
		if(linkin)
			globalscope->imported = listinsert(nil, globalscope->imported, scope);
		return;
	}
	if(listfind(opening, file)){
		warn(line, "circular inclusion of '%s'", file);
		return;
	}

	// Create a new scope for this module
	// linked to the real globalscope.
	scope = mkScope(globalframe, nil, nil);
	scope->end = globalframe->vars;
	scope->imported = list1(nil, realglobalscope);

	// Push the current global scope and parse.
	old = globalscope;
	old->start = old->frame->vars;
	globalscope = scope;
	opening = listinsert(nil, opening, file);

	if(pushinput(file, basedir)){
		pusharena(&asyntax);
		if(yyparse() < 0)
			panic("yyparse %s", file);
		afree(poparena(&asyntax));
		popinput();
	}

	// Link new into old.
	if(linkin)
		old->imported = listinsert(nil, old->imported, scope);

	// Pop and re-push the old globalscope.
	globalscope = mkScope(old->frame, old, old->frame->vars);

	opening = listpop(opening);
	hashput(opened, file, scope);
}

void
loadextension(char *name)
{
	queuedextensions = listappend((Arena*)nil, queuedextensions, nameof(name));
}

static void
loadqueuedextensions(void)
{
	List<Name> *l;

	l = queuedextensions;
	queuedextensions = ZList;
	while(l){
		dorequire(ZL, l->x, 0);
		l = listpop(l);
	}
}

static void
dopatternxforms(Thread *t)
{
	List<PatternXForm*> *l;
	PatternXForm *xf;

	for(l=queuedpatternxforms; l; l=l->next){
		xf = l->x;
		astviewxform(t, xf);
		free(xf);
	}
	listfree(queuedpatternxforms);
	queuedpatternxforms = nil;
}

void
builtinexit(void *v)
{
	struct fnframe {
		int e;
	} *f = (fnframe*)v;
	
	doexit();
	exit(f->e);
}

// Install one builtin.
void
installbuiltindata(char *name, void *data, Type *type)
{
	Var *v;

	Typecheck tc(globalframe, globalscope);
	name = nameof(name);
	Name n = nameof(name);
	if((v = lookup(tc.scope_, n, 0)) != nil)
		fprint(2, "warning: redefining %s\n", name);
	else{
		v = declare(&tc, ZL, n, type);
		assert(v);
	}
	memmove(globaldata()+v->off, data, sizeoftype(type));
}

void
installbuiltin(char *name,  void (*cfn)(void*), Type *type)
{
	Fn *fn;
	Name n = nameof(name);
	fn = mkbuiltinfn(n, cfn, type);
	installbuiltindata(n, &fn, type);
}

void
installbuiltintype(char *name, Type *type)
{
	Typecheck tc(globalframe, globalscope);
	declaretype(&tc, ZL, nameof(name), &type);
}

