// Type constructors.
//
// Arranged so that there is only one instance of any given type,
// so that type equality can be implemented as a pointer comparison.
// 
// (These types are for the zeta language, not for the language that
// might be compiled by a compiler written in zeta.)

#include "grammar.h"

#define listinsert(x, y) listinsert((Arena*)nil, x, y)
#define listcopy(x) listcopy((Arena*)nil, x)
#define listappend(x, y) listappend((Arena*)nil, x, y)
#define list1(x) list1((Arena*)nil, x)

Type*
typevoid(void)
{
	static Type *t;

	if(t == nil)
		t = typetuple(ZList);
	return t;
}

Type*
typedontcare(void)
{
	static Type *t;

	if(t == nil)
		t = mkTypeDontCare();
	return t;
}

Type*
typebool(void)
{
	static Type *t;
	
	if(t == nil)
		t = mkTypeBool();
	return t;
}

Type*
typeint(void)
{
	static Type *t;
	
	if(t == nil)
		t = mkTypeInt();
	return t;
}

Type*
typefloat(void)
{
	static Type *t;

	if(t == nil)
		t = mkTypeFloat();
	return t;
}

Type*
typestring(void)
{
	static Type *t;
	
	if(t == nil)
		t = mkTypeString();
	return t;
}

Type*
typeptr(Type *left)
{
	static Hash<Type*,Type*> *all;
	Type *t;
	
	if(all == nil)
		all = mkhash<Type*,Type*>();
	t = hashget(all, left);
	if(t == nil){
		t = mkTypePtr(left);
		hashput(all, left, t);
	}
	return t;
}

Type*
typearray(Type *left)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	if(left && isabstracttype(left)){
		error(ZL, "abstract element type in array(%T)", left);
		return nil;
	}

	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left)
			return t;
	}
	
	t = mkTypeArray(left);
	all = listinsert(all, t);
	return t;
}

Type*
typemap(Type *left, Type *right)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	if(left)
		switch(left->op){
		case TypeString:
		case TypeInt:
		case TypePtr:
			break;
		default:
			error(ZL, "bad index in map(%T, %T)", left, right);
			return nil;
		}

	if(right && isabstracttype(right)){
		error(ZL, "abstract element type %T in map(%T, %T)", left, right);
		return nil;
	}

	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left && t->right == right)
			return t;
	}
	
	t = mkTypeMap(left, right);
	all = listinsert(all, t);
	return t;
}

Type*
typemapiter(Type *left, Type *right)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left && t->right == right)
			return t;
	}
	
	t = mkTypeMapIter(left, right);
	all = listinsert(all, t);
	return t;
}

int
isassoclist(Type *type, Type **key, Type **val)
{
	Type *t, *k;

	if(type->op == TypeList && type->left->op == TypeTuple){
		t = type->left;
		k = listpeek(t->types);
		if(listlen(t->types) == 2)
			switch(k->op){
			case TypeString:
			case TypeInt:
			case TypePtr:
				if(key)
					*key = k;
				if(val)
					*val = t->types->next->x;
				return 1;
			default:
				error(ZL, "bad index for associative list%T", t);
				return 0;
			}
	}
	return 0;
}

Type*
typelist(Type *left)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	if(left && isabstracttype(left)){
		error(ZL, "abstract element type in list(%T)", left);
		return nil;
	}
	
	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left)
			return t;
	}
	
	t = mkTypeList(left);
	all = listinsert(all, t);
	return t;
}

Type*
typesubname(Type *_supername, Name name)
{
	static Hash<void*,List<Type*>*> *h;
	List<Type*> *l, *all;
	Type *t;
	void *key;
	
	assert(_supername);

	if(h == nil)
		h = mkhash<void*,List<Type*>*>();
	key = (void*)((uintptr)(char*)name ^ ((uintptr)_supername >> 3));
	for(l=all=hashget(h, key); l; l=l->next){
		t = l->x;
		if(t->_supername == _supername && t->name == name)
			return t;
	}

	t = mkTypeSubname(_supername, name);
	all = listinsert(all, t);
	hashput(h, key, all);

	// ugly hack: Sym.Fn is a subtype of Sym.Var, but we want Sym.Fn
	// to resolve correctly.  So for Sym.Var,
	//  ._supername = Sym
	//  .super      = Sym.Fn
	// The usual case is that (._supername == .super), so we fill
	// that in below.  It will be corrected when necessary.
	t->super = _supername;

	// Lazy resolution of grammar types.
	if(_supername->op == TypeSyntax || _supername->op == TypeUnion)
		resolvesyntaxsubtype(name, _supername, t);
	return t;
}

Type*
typefn(List<Type*> *argtypes, Type *rettype)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;
	
	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == rettype && listeq(argtypes, t->types))
			return t;
	}
	
	t = mkTypeFn(listcopy(argtypes), rettype);
	all = listinsert(all, t);
	return t;
}

Type*
typefnargs(List<Fnarg*> *fnargs, Type *rettype)
{
	Type *t;
	
	List<Type*> *l = ZList;
	for(List<Fnarg*> *ll=fnargs; ll; ll=ll->next)
		l = listappend(l, (ll->x)->type);
	t = typefn(l, rettype);
	listfree(l);
	return t;
}

Type*
typefn2(Type *in, Type *rettype)
{
	List<Type*> *l;
	Type *t;

	if(in == nil)
		return typefn(nil, rettype);
	if(in->op == TypeTuple)
		return typefn(in->types, rettype);
	l = list1(in);
	t = typefn(l, rettype);
	listfree(l);
	return t;
}

Type*
typetuple(List<Type*> *types)
{
	static List<Type*> *all;
	List<Type*> *l, *ts;
	Type *t;
	
	if(types != nil && types->next == nil)
		return types->x;

	for(l=all; l; l=l->next){
		t = l->x;
		if(listeq(types, t->types))
			return t;
	}

	ts = listcopy(types);
	t = mkTypeTuple(ts, ts);
	all = listinsert(all, t);
	return t;
}

Type*
typeunion(List<Type*> *types, struct Gram *grammar)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t, *tt;
	
	if(types != nil && types->next == nil)
		return types->x;
	assert(grammar);

	// must be a union of grammar types with the same family (could have a wildcard)
	for(l=types; l; l=l->next){
		t = l->x;
		switch(t->op){
		case TypeSyntax:
			assert(t->grammar == grammar);
			assert(t->super == nil);
			continue;
		case TypeDontCare:
			continue;
		default:
			panic("non syntax in TypeUnion: %T", t);
		}
	}

	// sort and remove dups
	types = listcopy(types);
	types = listsort(types, (int(*)(Type*,Type*))nil);
	types = listremovedups(types, (List<Type*>**)nil, (int(*)(Type*,Type*))nil);

	// remove all grammars t for which exists tt. (t != tt && t <: tt)
	List<Type*> **p;
	for(p=&types; *p; ){
		t = (*p)->x;
		if(t->op == TypeDontCare)
			goto continue_t;
		for(l=types; l; l=l->next){
			tt = l->x;
			if(t == tt || tt->op == TypeDontCare)
				continue;
			if(issubtypegrammar(tt, t)){
				l = *p;
				*p = (*p)->next;
				l->next = nil;
				listfree(l);
				goto next_t;
			}
		}
	continue_t:;
		p = &((*p)->next);
	next_t:;
	}

	// check again for a singleton
	assert(types);
	if(types != nil && types->next == nil)
		return types->x;

	// check if it exists
	for(l=all; l; l=l->next){
		t = l->x;
		if(listeq(types, t->types))
			return t;
	}

	// new type
	t = mkTypeUnion(types, grammar);
	//fprint(2, "U: %T\n", t);
	all = listinsert(all, t);
	adduniongrammar(t, grammar);
	return t;
}

Type*
typetupleexprs(const vector<Expr*>& exprs)
{
	Type *t;
	List<Type*> *l;

	l = ZList;
	for(int i=0; i<exprs.size(); i++)
		l = listappend(l, exprs[i]->type_);
	t = typetuple(l);
	listfree(l);
	return t;
}

Type*
typeforall(List<Type*> *typevars, Type *left)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	// sanity check
	for(l=typevars; l; l=l->next){
		t = l->x;
		assert(t->op == TypeVar);
	}

	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left && listeq(typevars, t->typevars))
			return t;
	}
	
	t = mkTypeForall(listcopy(typevars), left);
	all = listinsert(all, t);
	return t;
}

Type*
typesubst(List<Type*> *types, List<Type*> *typevars, Type *left)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;

	// sanity check
	assert(listlen(types) == listlen(typevars));

	for(l=all; l; l=l->next){
		t = l->x;
		if(t->left == left && listeq(t->types, types) && listeq(t->typevars, typevars))
			return t;
	}

	t = mkTypeSubst(listcopy(types), listcopy(typevars), left);
	all = listinsert(all, t);
	return t;
}

Type*
typeunresolved(Scope *scope, Name name)
{
	static List<Type*> *all;
	List<Type*> *l;
	Type *t;
	
	for(l=all; l; l=l->next){
		t = l->x;
		if(t->declscope == scope && t->name == name)
			return t;
	}
	
	t = mkTypeUnresolved(scope, name);
	all = listinsert(all, t);
	return t;
}

List<Type*>*
names2typevars(List<Name> *names)
{
	List<Type*> *l;
	l = ZList;
	for(; names; names=names->next)
		l = listappend(l, mkTypeVar(names->x));
	return l;
}

// Builtins
#define declbuiltin(name) \
	Type* _ ## name; /* set in builtin.c */ \
	Type* name(void) \
	{ \
		if(_ ## name == nil) \
			panic(#name); \
		return _ ## name; \
	}

declbuiltin(typeast)
declbuiltin(typeastmerge)
declbuiltin(typeastslot)
declbuiltin(typeaststring)
declbuiltin(typeline)
declbuiltin(typeframe)
declbuiltin(typescope)
declbuiltin(typesym)

#undef declbuiltin

// Compute the size of a type (how many contiguous bytes to allocate).
int
sizeoftype(Type *t)
{
	List<Type*> *l;
	int n;

	if(t == nil)
		panic("sizeoftype: nil");
	switch(t->op){
	case TypeInt:
	case TypeBool:
		return sizeof(int);
	case TypeFloat:
		return sizeof(double);
	case TypeString:
	case TypeList:
	case TypeFn:
	case TypePtr:
	case TypeVar:
	case TypeArray:
	case TypeMap:
		return sizeof(void*);
	case TypeMapIter:
		return sizeof(MapIter) + sizeoftype(t->right);
	case TypeTuple:
		n = 0;
		for(l=t->types; l; l=l->next)
			n += sizeoftype(l->x);
		return n;
	case TypeStruct:
		n = 0;
		// See dat.h for a data layout diagram.
		if(t->extends)
			n = offsetof(ExtStruct, v);
		else if(t->super){
			n = sizeoftype(t->super);
			assert(t->reserved == 0);
		}
		else
			n = structdataoffset(t);
		for(l=t->types; l; l=l->next)
			n += sizeoftype(l->x);
		return n;
	case TypeSyntax:
	case TypeUnion:
		return sizeoftype(t->structrep);
	case TypeForall:
	case TypeSubst:
		return sizeoftype(t->left);
	case TypeName:
		panic("sizeoftype: Name type %T", t);
	case TypeSubname:
		panic("sizeoftype: Subname type %T", t);
	case TypeUnresolved:
		panic("sizeoftype: Unresolved type %T", t);
	case TypeDontCare:
		panic("sizeoftype: DontCare");
	}
	panic("sizeoftype: ?%d?", t->op);
	return 0;
}

// Translate TypeAst to TypeStruct (eg, G.sym.rule --> Ast.sym.rule),
// and unfold TypeForall and TypeSubst types to get a TypeStruct.
static Type*
unfoldstruct(Type *t)
{
	assert(t);
	if(t->op == TypeSyntax || t->op == TypeUnion)
		t = t->structrep;
	else{
		if(t && t->op == TypeSubst)
			t = t->left;
		if(t && t->op == TypeForall)
			t = t->left;
	}
	assert(t && t->op == TypeStruct);
	return t;
}

// Compute the offset of the first data field for a struct.
int
structdataoffset(Type *t)
{
	int n;

	t = unfoldstruct(t);
	// Extension struct
	if(t->extends)
		return offsetof(ExtStruct, v);
	// Tagged or normal struct
	while(t->super)
		t = t->super;
	n = t->reserved;
	if(t->extensible)
		n += sizeof(void*);
	if(t->discriminated)
		n += sizeof(void*);
	return n;
}

// Compute the offset of the extension ptr for an extensible struct.
int
extptroffset(Type *t)
{
	t = unfoldstruct(t);
	while(t->super)
		t = t->super;
	assert(t->extensible);
	return t->reserved;
}

// Compute the offset of the tag type ptr for a discriminated struct.
int
tagtypeoffset(Type *t)
{
	int n;

	t = unfoldstruct(t);
	while(t->super)
		t = t->super;
	assert(t->discriminated);
	n = t->reserved;
	if(t->extensible)
		n += sizeof(void*);
	return n;
}

static int typefmtlisp(Fmt*, Type*);

static int
typestructfmt(Fmt *fmt, Type *s, Type *t)
{
	if(s == nil)
		return 0;
	if(s->super){
		if(s != lookuptagowner(t))
			typestructfmt(fmt, s->super, t);
		else
			typestructfmt(fmt, s->super, s);
	}
	const char *sep = (s == t) ? "" : ".";
	if(s->op == TypeUnion)
		return fmtprint(fmt, "%T%s", s, sep);
	else
		return fmtprint(fmt, "%s%s", s->name, sep);
}

int
typefmt(Fmt *fmt)
{
	Type *t;
	List<Type*> *l;
	char *sep;
	
	t = va_arg(fmt->args, Type*);
	if(t == nil)
		return fmtstrcpy(fmt, "???");
	
	if(fmt->flags&FmtSharp)
		return typefmtlisp(fmt, t);

	switch(t->op){
	case TypeBool:
		return fmtstrcpy(fmt, "bool");
	case TypeInt:
		return fmtstrcpy(fmt, "int");
	case TypeFloat:
		return fmtstrcpy(fmt, "float");
	case TypeString:
		return fmtstrcpy(fmt, "string");
	case TypeList:
		// TO DO: Parenthesize %T if needed.
		return fmtprint(fmt, "list %T", t->left);
	case TypeFn:
		// TO DO: Parenthesize %T if needed.
		if(t->types && !t->types->next)
			return fmtprint(fmt, "%T -> %T", t->types->x, t->left);
		fmtprint(fmt, "(");
		sep = "";
		for(l=t->types; l; l=l->next){
			fmtprint(fmt, "%s%T", sep, l->x);
			sep = ", ";
		}
		return fmtprint(fmt, ") -> %T", t->left);
	case TypeTuple:
	case TypeUnion:
		fmtprint(fmt, "(");
		sep = "";
		for(l=t->types; l; l=l->next){
			fmtprint(fmt, "%s%T", sep, l->x);
			if(t->op == TypeTuple)
				sep = ", ";
			else
				sep = " + ";
		}
		return fmtprint(fmt, ")");
	case TypeStruct:
		fmtprint(fmt, "struct ");
		return typestructfmt(fmt, t, t);
	case TypeSyntax:
		fmtprint(fmt, "syntax ");
		return typestructfmt(fmt, t, t);
	case TypePtr:
		return fmtprint(fmt, "ptr %T", t->left);
	case TypeName:
		return fmtprint(fmt, "(unresolved name %s)", t->name);
	case TypeSubname:
		return fmtprint(fmt, "(unresolved %T.%s)", t->super, t->name);
	case TypeUnresolved:
		return fmtprint(fmt, "(unresolved %s)", t->name);
	case TypeArray:
		return fmtprint(fmt, "array %T", t->left);
	case TypeMap:
		return fmtprint(fmt, "map(%T, %T)", t->left, t->right);
	case TypeMapIter:
		return fmtprint(fmt, "mapiter(%T, %T)", t->left, t->right);
	case TypeDontCare:
		return fmtprint(fmt, "_");
	case TypeVar:
		return fmtprint(fmt, "<%s>", t->name);
	case TypeSubst:
		fmtprint(fmt, "<");
		sep = "";
		for(l=t->types; l; l=l->next){
			fmtprint(fmt, "%s%T", sep, l->x);
			sep = ", ";
		}
		fmtprint(fmt, "> / <");
		// fallthrough
	case TypeForall:
		if(t->op == TypeForall)
			fmtprint(fmt, "forall <");
		sep = "";
		for(l=t->typevars; l; l=l->next){
			fmtprint(fmt, "%s%s", sep, l->x ? (char*)(l->x)->name : "???");
			sep = ", ";
		}
		return fmtprint(fmt, ">. %T", t->left);
	}
	return fmtprint(fmt, "?%d?", t->op);
}

static int
typefmtlisp(Fmt *fmt, Type *t)
{
	if(t == nil)
		return fmtstrcpy(fmt, "nil");
	
	fmtprint(fmt, "(%s", strTypeOp(t->op));
	if(t->name)
		fmtprint(fmt, " %s", t->name);
	if(t->left){
		fmtprint(fmt, " ");
		typefmtlisp(fmt, t->left);
	}
	if(t->right){
		fmtprint(fmt, " ");
		typefmtlisp(fmt, t->right);
	}
	if(t->super){
		fmtprint(fmt, " ");
		typefmtlisp(fmt, t->super);
	}
	return fmtprint(fmt, ")");
}

// Helper for maptypechildren.  Copy a list of types and apply xform
// to each type in the list.  Set *ch to 1 if anything in the list is
// changed.
static List<Type*>*
maptypelist(void *s, List<Type*> *ts, int *ch,
	    Type* (*xform)(void*, Type *))
{
	Type *sub;
	List<Type*> *l, *types;

	types = listcopy(ts);
	for(l=types; l; l=l->next){
		sub = xform(s, l->x);
		if(sub){
			l->x = sub;
			*ch = 1;
		}
	}
	return types;
}

// Run a transformer and return the type if it didn't change.
static Type*
filterxform(void *s, Type *t, Type* (*xform)(void*, Type *))
{
	Type *x;
	x = xform(s, t);
	return x ? x : t;
}

// Map a transformer function across a Type.  The transformer function
// will be called with each child of the given type.  The transformer
// function should return either a type to replace the child type, or
// nil to indicate no change.
//
// Returns nil if the type is not changed by the transformer (ie, all
// calls to the transformer returned nil).  Otherwise, returns a type
// with the same op as the input type, but with changes to the
// children where the transformer return non-nil.
Type*
_maptypechildren(void *s, Type *t, Type* (*xform)(void*, Type *))
{
	List<Type*> *types;
	Type *sub, *out, *sub1;
	Var *var;
	int ch;

	out = nil;
	ch = 0;

	if(t == nil)
		return nil;
	switch(t->op){
	case TypeBool:
	case TypeInt:
	case TypeFloat:
	case TypeString:
	case TypeUnresolved:
	case TypeDontCare:
	case TypeSyntax:
	case TypeVar:
		break;
	case TypeList:
		sub = xform(s, t->left);
		if(sub)
			out = typelist(sub);
		break;
	case TypeFn:
		types = maptypelist(s, t->types, &ch, xform);
		sub = xform(s, t->left);
		if(ch || sub){
			if(!sub)
				sub = t->left;
			out = typefn(types, sub);
		}
		listfree(types);
		break;
	case TypeTuple:
		types = maptypelist(s, t->types, &ch, xform);
		if(ch)
			out = typetuple(types);
		listfree(types);
		break;
	case TypeUnion:
		types = maptypelist(s, t->types, &ch, xform);
		if(ch)
			out = typeunion(types, t->grammar);
		listfree(types);
		break;
	case TypeStruct:
		// This *is* the canonical struct, so modify it in-place
		for(List<Type*> *l=t->types; l; l=l->next)
			l->x = filterxform(s, l->x, xform);
		for(List<Type*> *l=t->ctorargs; l; l=l->next)
			l->x = filterxform(s, l->x, xform);
		for(List<Var*> *l=t->vars; l; l=l->next){
			var = l->x;
			var->type = filterxform(s, var->type, xform);
		}
		for(List<Type*> *l=t->tags; l; l=l->next)
			xform(s, l->x);
		for(List<Type*> *l=t->extensions; l; l=l->next)
			xform(s, l->x);
		break;
	case TypePtr:
		// Don't recurse on structs to prevent infinite recursion.
		if(t->left && t->left->op == TypeStruct)
			break;
		sub = xform(s, t->left);
		if(sub)
			out = typeptr(sub);
		break;
	
	case TypeSubname:
		sub = xform(s, t->_supername);
		if(sub)
			out = typesubname(sub, t->name);
		break;
	
	case TypeArray:
		sub = xform(s, t->left);
		if(sub)
			out = typearray(sub);
		break;
	
	case TypeMap:
		sub = xform(s, t->left);
		sub1 = xform(s, t->right);
		if(sub || sub1){
			if(sub == nil)
				sub = t->left;
			if(sub1 == nil)
				sub1 = t->right;
			out = typemap(sub, sub1);
		}
		break;
	
	case TypeMapIter:
		sub = xform(s, t->left);
		sub1 = xform(s, t->right);
		if(sub || sub1){
			if(sub == nil)
				sub = t->left;
			if(sub1 == nil)
				sub1 = t->right;
			out = typemapiter(sub, sub1);
		}
		break;

	case TypeName:
		types = maptypelist(s, t->types, &ch, xform);
		if(ch)
			out = mkTypeName(t->name, types);
		break;

	case TypeForall:
		// Don't recurse on structs to prevent infinite recursion.
		if(t->left && t->left->op == TypeStruct)
			break;
		sub = xform(s, t->left);
		if(sub)
			out = typeforall(t->typevars, sub);
		break;

	case TypeSubst:
		// Don't recurse on structs to prevent infinite recursion.
		if(t->left && t->left->op != TypeStruct)
			break;
		sub = xform(s, t->left);
		types = maptypelist(s, t->types, &ch, xform);
		if(ch || sub){
			if(!sub)
				sub = t->left;
			out = typesubst(types, t->typevars, sub);
		}
		listfree(types);
		break;
	}

	return out;
}

// Can a value of type t1 be completed as a value of type t?
// If so, the caller should change t1 to t.
int
shouldcompletetype(Type *t, Type *t1)
{
	List<Type*> *l, *l1;

	if(t == nil || t1 == nil)
		return 0;
	if(t->op == TypeMap && t1->op == TypeMap && t1->left == nil && t1->right == nil)
		return 1;
	if(t->op == TypeList && t1->op == TypePtr && t1->left == nil)
		return 1;
	if(t->op == TypeList && t1->op == TypeList && t1->left == nil)
		return 1;
	if(t->op == TypeArray && t1->op == TypeArray && t1->left == nil)
		return 1;
	if(t->op == TypeTuple && t1->op == TypeTuple){
		for(l=t->types, l1=t1->types; l && l1; l=l->next, l1=l1->next)
			if(l->x != l1->x && !shouldcompletetype(l->x, l1->x))
				return 0;
		if(l || l1)
			return 0;
		return 1;
	}
	return 0;
}

// Is this type <: ptr Ast?
int
isptrast(Type *type)
{
	return type && type->op == TypePtr && type->left
		  && (type->left->op == TypeSyntax || type->left->op == TypeUnion);
}

// Is a type abstract?  If so, it cannot be constructed (either
// on the stack, with new, or as an element type in a struct,
// array, list, tuple, or map).
int
isabstracttype(Type *type)
{
	// Consider 'no type' abstract.
	if(type == nil)
		return 1;
	// The base struct of a discriminated union is abstract.
	// Note that the `discriminated' flag is not set in the tags.
	if(type->op == TypeStruct && type->discriminated)
		return 1;
	if(type->op == TypeSyntax || type->op == TypeUnion)
		return 1;
	return 0;
}

// Can a value of type sub be used as a value of type super?  In
// domain parlance, this is the $\supseteq$ operator
static int
_issubtype(Type *super, Type *sub, int dolist, int doptr)
{
	if(sub == nil || super == nil)
		return 0;
	if(sub == super)
		return 1;
	// Syntax subtyping
	if(sub->op == TypeSyntax || sub->op == TypeUnion){
		if(super->op != TypeSyntax && super->op != TypeUnion)
			return 0;
		return issubtypesyntax(super, sub);
	}
	// Struct subtyping
	if(sub->op == TypeStruct){
		if(super->op != TypeStruct)
			return 0;
		return isverticalsubtype(super, sub);
	}
	// Tuple subtyping
	// This is always safe because tuples are immutable
	if(sub->op == TypeTuple && super->op == TypeTuple){
		List<Type*> *la, *lb;
		for(la=super->types, lb=sub->types; la && lb; la=la->next, lb=lb->next){
			if(!_issubtype(la->x, lb->x, 1, 1))
				return 0;
		}
		return !la && !lb;
	}
	// Function subtyping.
	// Args are contravariant, return value is covariant.
	if(super->op == TypeFn && sub->op == TypeFn){
		List<Type*> *la, *lb;
		for(la=super->types, lb=sub->types; la && lb; la=la->next, lb=lb->next){
			if(!_issubtype(lb->x, la->x, 1, 1))
				return 0;
		}
		if(la || lb)
			return 0;
		if(!_issubtype(super->left, sub->left, 1, 1))
			return 0;
		return 1;
	}
	// ptr(nil) <: string
	// ptr(nil) <: fn
	// ptr(nil) <: ptr(T)
	// ptr(nil) <: <T>
	if(sub->op == TypePtr && sub->left == nil){
		if(super->op == TypeString || super->op == TypeFn
		|| super->op == TypePtr || super->op == TypeVar)
			return 1;
	}
	// Typevar subtyping (this probably isn't always sound!!!!!!)
	if(sub->op == TypeSubst && super->op == TypeSubst && sub->left == super->left){
		List<Type*> *la, *lb;
		for(la=super->types, lb=sub->types; la && lb; la=la->next, lb=lb->next){
			if(!_issubtype(la->x, lb->x, 1, 1))
				return 0;
		}
		return !la && !lb;
	}
	// Allow "list A" <: "list B" if A <: B and A <: ptr(Ast).
	// This is VERY unsafe when used with mutable lists, so be careful!
	// Not having this property, though, makes dealing with Ast lists
	// very annoying.  Maybe it should be restricted to lists of
	// immutable values (eg, Asts), only?
	if(dolist && super->op == TypeList && sub->op == TypeList)
		return _issubtype(super->left, sub->left, 0, 1);
	// Allow "map (A,B): <: "map (A,C)" if B <: C
	if(dolist && super->op == TypeMap && sub->op == TypeMap && super->left == sub->left)
		return _issubtype(super->right, sub->right, 0, 1);
	// Allow "array A" <: "array B" if A <: B
	if(dolist && super->op == TypeArray && sub->op == TypeArray)
		return _issubtype(super->left, sub->left, 0, 1);
	// Allow "ptr A" <: "ptr B" if A <: B.
	// This is safe only when A and B are immutable, but is convenient.
	// Consider:
	//    a: ptr A;  // A <: B
	//    c: ptr C;  // C <: B
	//    b: ptr B;
	//    b = a;     // safe by this rule
	//    *b = *c;   // safe by normal subtyping, but a is borked!
	if(!doptr || super->op != TypePtr || sub->op != TypePtr)
		return 0;
	// Note that "ptr ptr A" <: "ptr ptr B" is never legal.
	// This is unsafe even when A and B are immutable values.
	return _issubtype(super->left, sub->left, 0, 0);
}

int
issubtype(Type *super, Type *sub)
{
	return _issubtype(super, sub, 1, 1);
}

int
issubtypesyntax(Type *super, Type *sub)
{
	assert(sub   && (sub->op   == TypeSyntax || sub->op   == TypeUnion));
	assert(super && (super->op == TypeSyntax || super->op == TypeUnion));

	// Common case.
	if(super == typeast())
		return 1;
	if(sub == typeast())
		return 0;

	// Both types must belong to the same grammar family.
	if(sub->grammar && super->grammar && (sub->grammar != super->grammar))
		return 0;

	// Another common: "ast ~ (F+?)", eg, (ast ~ C) in xoc.
	// The type (G+?) is the Top in intra-family subtyping.
	if(super->grammar && super == super->grammar->wildroottype)
		return 1;

	// Unfold the view mirror.
	if(sub->mirrors)
		sub = sub->mirrors;
	if(super->mirrors)
		super = super->mirrors;

	int nsub = 0;
	int nsuper = 0;
	Type *t;

	// Trivial vertical subtyping.
	for(t=sub; t; nsub++, t=t->super)
		if(t == super)
			return 1;
	for(t=super; t; nsuper++, t=t->super)
		;
	// nsub and nsuper describe the "level" of the type:
	// Level 1:      Ast           G       // mught be a TypeUnion at this level
	// Level 2:   Ast.special    G.sym
	// Level 3:                G.sym.rule
	if(nsuper > nsub)
		return 0;
	// Check the reserved symnames (slot & merge).
	// These syms can exist in one of four places:
	//   Ast.slot   G.slot           (level == 2)
	//              G.sym.slot       (level == 3)
	//              G.sym.rule.slot  (level == 4; merge makes no sense here)
	if(super->name == Name::slot || super->name == Name::merge){
		if(sub->name != super->name)
			return 0;
		sub   = sub->super;
		super = super->super;
		assert(sub && super);
		nsub--;
		nsuper--;
	}
	else if(sub->name == Name::slot || sub->name == Name::merge){
		sub = sub->super;
		assert(sub);
		nsub--;
	}
	// Bring sub up to the level of super.
	for(; nsub > nsuper; nsub--)
		sub = sub->super;
	// Both types must have the same name for levels 2,3
	for(; nsub > 1; nsub--){
		if(sub->name != super->name)
			return 0;
		sub = sub->super;
		super = super->super;
	}
	// Everything <: Ast
	if(super == typeast())
		return 1;
	if(sub == typeast())
		return 0;
	// Grammar subtyping.  We already know both grammars belong to
	// the same family.  The idea is that A <: B if A includes all
	// CFG productions in B.  The basic formula is:
	//   forall g in sub. exists h in super. g <: h
	// With the additional rule for wildcards:
	//    everything <: {... + ?}
	List<Type*> _subs = { sub, nil };
	List<Type*> _supers = { super, nil};
	List<Type*> *subs   = (sub->op   == TypeUnion) ? sub->types   : &_subs;
	List<Type*> *supers = (super->op == TypeUnion) ? super->types : &_supers;
	List<Type*> *a, *b;
	int needswild=0;

	for(a=subs; a; a=a->next){
		Type *ta = a->x;
		// if sub has a wildcard, super must have one also
		if(ta->op == TypeDontCare){
			needswild = 1;
			continue;
		}
		for(b=supers; b; b=b->next){
			Type *tb = b->x;
			// wildcards are at the top
			if(tb->op == TypeDontCare)
				return 1;
			if(issubtypegrammar(tb, ta))
				goto next_a;
		}
		return 0;
	next_a:;
	}

	return !needswild;
}

int
issubtypegrammar(Type *super, Type *sub)
{
	assert(sub   && sub->op   == TypeSyntax && sub->super   == nil);
	assert(super && super->op == TypeSyntax && super->super == nil);

	if(sub->grammar != super->grammar)
		return 0;
	// DFS for a match.
	if(sub == super)
		return 1;
	List<Type*> *a;
	for(a=sub->supergrammars; a; a=a->next)
		if(issubtypegrammar(super, a->x))
			return 1;
	return 0;
}

// Is sub <: super by vertical struct subtyping?
int
isverticalsubtype(Type *super, Type *sub)
{
	for(; sub; sub=sub->super)
		if(sub == super)
			return 1;
	return 0;
}

// Compute the least common supertype of two types.
// Return nil if the types cannot be joined.
static Type*
_subtypejoin(Type *a, Type *b, int dolist, int doptr)
{
	Type *t, *ta, *tb;

	if(a == nil || b == nil)
		return nil;
	if(a == b)
		return a;
	// Tuple subtyping
	if(a->op == TypeTuple && b->op && TypeTuple){
		List<Type*> *la, *lb, *l;
		l = ZList;
		for(la=a->types, lb=b->types; la && lb; la=la->next, lb=lb->next){
			t = _subtypejoin(la->x, lb->x, 1, 1);
			if(!t){
				listfree(l);
				return nil;
			}
			l = listappend(l, t);
		}
		if(la || lb){
			listfree(l);
			return nil;
		}
		return typetuple(l);
	}
	// cannot join struct types unless under a pointer
	if(!doptr){
		if(a->op == TypeStruct && b->op == TypeStruct){
			for(ta=a; ta; ta=ta->super)
				for(tb=b; tb; tb=tb->super)
					if(ta == tb)
						return ta;
			return nil;
		}
		if((a->op == TypeSyntax || a->op == TypeUnion)
		&& (b->op == TypeSyntax || b->op == TypeUnion)){
			for(ta=a; ta; ta=ta->super)
				for(tb=b; tb; tb=tb->super)
					if(ta == tb)
						return ta;
			Type *ag = gramsupertype(a);
			Type *bg = gramsupertype(b);
			Type *g;
			if(ag && bg && (g = joingrammartypes(ag, bg))){
				// Possibly narrow to a sym/rule.
				Type *as = symsupertype(a);
				Type *bs = symsupertype(a);
				Type *ar = rulesupertype(a);
				Type *br = rulesupertype(b);
				if(as && bs && as->name == bs->name){
					static Name slot, merge;
					if(!slot){
						slot = nameof("slot");
						merge = nameof("merge");
					}
					// If the join is Ast, this must be a special symbol
					if(g == typeast() && as->name != slot && as->name != merge)
						return g;
					t = lookuptag(g, as->name, nil);
					assert(t);
					if(ar && br && ar->name == br->name){
						t = lookuptag(t, ar->name, nil);
						assert(t);
					}
					return t;
				}
				return g;
			}
			return nil;
		}
	}
	// ptr(nil) <: string
	// ptr(nil) <: fn
	// ptr(nil) <: ptr(T)
	if(a->op == TypePtr && a->left == nil){
		if(b->op == TypeString || b->op == TypeFn || b->op == TypePtr)
			return b;
	}
	if(b->op == TypePtr && b->left == nil){
		if(a->op == TypeString || a->op == TypeFn || a->op == TypePtr)
			return a;
	}
	// Typevar subtyping (this probably isn't always sound!!!!!!)
	if(a->op == TypeSubst && b->op == TypeSubst && a->left == b->left){
		List<Type*> *la, *lb, *l;
		l = ZList;
		for(la=a->types, lb=b->types; la && lb; la=la->next, lb=lb->next){
			t = _subtypejoin(la->x, lb->x, 1, 1);
			if(!t){
				listfree(l);
				return nil;
			}
			l = listappend(l, t);
		}
		if(la || lb){
			listfree(l);
			return nil;
		}
		return typesubst(l, a->typevars, a->left);
	}
	// "list A" <: "list B" if A <: B
	// Unsafe! (see note in issubtype)
	if(dolist && a->op == TypeList && b->op == TypeList){
		t = _subtypejoin(a->left, b->left, 0, 1);
		if(t == nil)
			return nil;
		return typelist(t);
	}
	// "ptr A" <: "ptr B" if A <: B
	// Unsafe! (see note in issubtype)
	if(!doptr || a->op != TypePtr || b->op != TypePtr)
		return nil;
	t = _subtypejoin(a->left, b->left, 0, 0);
	if(t == nil)
		return nil;
	return typeptr(t);
}

Type*
subtypejoin(Type *a, Type *b)
{
	return _subtypejoin(a, b, 1, 1);
}

// Compute the least-upper-bound of two grammar types.
Type*
joingrammartypes(Type *a, Type *b)
{
	if(a == b)
		return a;
	if(!a) return b;
	if(!b) return a;
	assert(a->op == TypeSyntax || a->op == TypeUnion);
	assert(b->op == TypeSyntax || b->op == TypeUnion);
	// Must have be from the same grammar family, else we join to Ast
	if(a->grammar != b->grammar)
		return typeast();
	// Just union everything together and typeunion() will canonicalize.
	List<Type*> *alist = (a->op == TypeSyntax) ? list1(a) : listcopy(a->types);
	List<Type*> *blist = (b->op == TypeSyntax) ? list1(b) : listcopy(b->types);
	return typeunion(listconcat(alist, blist), a->grammar);
}

typedef struct IsResolvedState IsResolvedState;
struct IsResolvedState
{
	int isUnresolved;
	Type **unresolved;
};

static Type*
isunresolved1(IsResolvedState *rs, Type *t)
{
	if(t == nil)
		return nil;

	if(rs->isUnresolved || (t && t->isResolved)) {
		// Don't bother traversing any further
		return nil;
	}

	if(t->op == TypeUnresolved) {
		rs->isUnresolved = 1;
		if(rs->unresolved)
			*(rs->unresolved) = t;
	} else if(t->op == TypePtr || t->op == TypeFn) {
		// We know the representation of pointer and function
		// types regardless of their children
		t->isResolved = 1;
	} else {
		// Assume that this type is resolved
		t->isResolved = 1;
		// Check the children
		maptypechildren(rs, t, isunresolved1);
		// Were any of the children unresolved?
		if(rs->isUnresolved) {
			// Our assumption was disproved
			t->isResolved = 0;
		}
	}
	return nil;
}

// Determine whether or not a type is resolved.  This recurses into
// the type to check for TypeUnresolved.  If an unresolved type is
// found, this returns 1 and (if unresolved is non-nil), points
// unresolved at the unresolved type.
//
// A type is "resolved" if and only if we know its representation (and
// thus, if and only if we can find its size).  Thus, for example,
// pointer types are always resolved because we don't need to know the
// representation of what they're pointing at to know their
// representation.
int
isunresolved(Type *t, Type **unresolved)
{
	IsResolvedState rs;
	rs.isUnresolved = 0;
	rs.unresolved = unresolved;

	if (unresolved)
		*unresolved = nil;

	isunresolved1(&rs, t);
	return rs.isUnresolved;
}

// ========================================================
// Type substitution

typedef struct SubstState SubstState;
struct SubstState
{
	Line line;
	List<Type*> *vars;
	List<Type*> *types;
	int hasfreevar;
};

static Type*
applytypesubst1(SubstState *s, Type *t)
{
	List<Type*> *lv, *lt;

	if(t == nil || t->op == TypeStruct)
		return nil;
	if(t->op == TypeVar){
		for(lv=s->vars, lt=s->types; lv; lv=lv->next, lt=lt->next)
			if(lv->x == t)
				return lt->x;
		return nil;
	}
	return maptypechildren(s, t, applytypesubst1);
}

Type*
applytypesubst(Type *t, List<Type*> *vars, List<Type*> *types)
{
	SubstState s;
	Type *out;

	assert(listlen(vars) == listlen(types));
	s.vars  = vars;
	s.types = types;

	out = applytypesubst1(&s, t);
	if(out)
		return out;
	return t;
}

// ========================================================
// Type unification.
// Attempt to unify 'have' with 'expected', where 'types'
// is the current substitution for 'vars'.  Assume that
// typevars from 'vars' do not appear in 'have' (only 'expected').

static int
canunifytypes1(SubstState *s, Type *have, Type *expected)
{
	List<Type*> *l, *ll;
	List<Type*> *lv, *lt;

	if(have == expected)
		return 1;
	if(!have || !expected)
		return 0;

	if(expected->op == TypeVar){
		for(lv=s->vars, lt=s->types; lv; lv=lv->next, lt=lt->next){
			if(lv->x != expected)
				continue;
			if(lt->x)
				return canunifytypes1(s, have, lt->x);
			// TODO: occurs check?
			lt->x = have;
			return 1;
		}
	}
	if(have->op != expected->op){
		if((have->op == TypeStruct     || have->op == TypeSyntax     || have->op == TypeUnion)
		&& (expected->op == TypeStruct || expected->op == TypeSyntax || expected->op == TypeUnion))
		{
			Type *t = subtypejoin(typeptr(have), typeptr(expected));
			if(t)
				return 1;
		}
		error(s->line, "cannot unify [%T] with [%T]", have, expected);
		return 0;
	}

	switch(have->op){
	case TypeMap:
	case TypeMapIter:
	case TypeForall:
		if(!canunifytypes1(s, have->right, expected->right))
			return 0;
		// fallthrough
	case TypeArray:
	case TypeList:
	case TypePtr:
		return canunifytypes1(s, have->left, expected->left);

	case TypeFn:
	case TypeSubst:
		if(!canunifytypes1(s, have->left, expected->left))
			return 0;
		// fallthrough
	case TypeTuple:
		if(listlen(have->types) != listlen(expected->types))
			return 0;
		for(l=have->types, ll=expected->types; l; l=l->next, ll=ll->next)
			if(!canunifytypes1(s, l->x, ll->x))
				return 0;
		return 1;

	default:
		return 1;
	}
}

int
canunifytypes(Line line, Type *have, Type *expected, List<Type*> *vars, List<Type*> *types)
{
	SubstState s;

	assert(listlen(vars) == listlen(types));
	s.line  = line;
	s.vars  = vars;
	s.types = types;
	return canunifytypes1(&s, have, expected);
}

// ========================================================
// TypeStruct operations
//
// Zeta structs have gotten a bit odd.  This is an attempt to
// explain the most confusing aspect of the current semantics.
//
// Structs actually have three orthagonal features:
// - Extensions
//     (obvious)
// - Subtyping
//     A struct can inherit the fields of another struct by
//     subtyping.  Type::super gives the direct parent of
//     a child struct.
// - Tags
//     A struct can live in the namespace of another struct
//     via 'tags'.  If struct "T" is a tag of struct "S", its
//     global name is "S.T", and T is a member of S->tags.
//     This is usually combined with subtyping, but the tag
//     tree is not necessarily the same as the subtyping tree.
//     For example:
//
//         struct Sym {
//            tag Var {};
//            tag Fndef : Var {};
//         }
//
//     This defines three structs: Sym, Sym.Var, and Sym.Fndef.
//     The trees are:
//        subtyping : Sym > Var > Fndef
//        tags      : Sym > {Var, Fndef}
//
//     Note that Fndef is a tag of Sym, but a direct subtype
//     of Var.  All user-defined structs act as above.

void
addstructtag(Type *s, Type *t)
{
	List<Type*> *l;

	assert(s && (s->op == TypeStruct || s->op == TypeSyntax || s->op == TypeUnion));
	assert(t && (t->op == TypeStruct || t->op == TypeSyntax));
	assert(s->tagshash);

	l = hashget(s->tagshash, t->name);
	l = listappend(l, t);
	hashput(s->tagshash, t->name, l);
	s->tags = listappend(s->tags, t);
}

static int globalattrsfixed = 0;
void
fixglobalattrs(void)
{
	// We've set all the global attributes, and won't be
	// adding anymore.  This is a paranoid safety mechanism.
	globalattrsfixed = 1;
}

Var*
addstructattr(Name name, Type *type, Type *host, int isbuiltin)
{
	Var *v;

	assert(type);
	assert(host);
	assert(host->op == TypeStruct);

	// Sanity check: all attributes in the globally-shared "Ast"
	// host should be builtins (eg, "prev" and "next").
	if(host->super == nil){
		assert(!globalattrsfixed);
		assert(isbuiltin);
	}

	v = mkVar(name, type);
	v->off = host->attrsize;
	v->isconst = 1;
	v->isbuiltinattr = isbuiltin;
	v->host = host;
	host->attrsize += sizeof(int);
	host->attrsize += sizeoftype(type);
	host->vars = listappend(host->vars, v);
	return v;
}

// Is scope 's' visible from scope 'from'?
static int
isvisiblescope(Scope *s, Scope *from)
{
	if(s == nil || from == nil || s == globalscope)
		return 1;
	for(; from; from=from->outer){
		if(from == s)
			return 1;
		for(List<Scope*> *l=from->imported; l; l=l->next)
			if(isvisiblescope(s, l->x))
				return 1;
	}
	return 0;
}

// Lookup a field in a struct.  If it must be accessed through
// an extension, `xtp' is set to the extension structure in
// which the field lives.  TypeSyntaxs are translated to their
// TypeStruct.
Var*
lookupstruct(Type *t, Name name, Scope *scope, Type **xtp)
{
	Var *v;
	Type *tt;

	t = unfoldstruct(t);
	assert(!t->extends);

	if(xtp)
		*xtp = nil;
	// Self
	for(List<Var*> *l=t->vars; l; l=l->next){
		v = l->x;
		if(!isvisiblescope(v->declscope, scope))
			continue;
		if(v->name == name)
			return v;
	}
	if(t->extensible){
		for(List<Type*> *ll=t->extensions; ll; ll=ll->next){
			tt = ll->x;
			if(!isvisiblescope(tt->declscope, scope))
				continue;
			for(List<Var*> *l=tt->vars; l; l=l->next){
				v = l->x;
				if(v->name == name){
					if(xtp)
						*xtp = tt;
					return v;
				}
			}
		}
	}
	// Super
	if(t->super)
		return lookupstruct(t->super, name, scope, xtp);
	return nil;
}

// Lookup a tag struct which is a direct child of struct 't'.
Type*
lookuptag(Type *t, Name name, Scope *scope)
{
	List<Type*> *tl;
	Type *tag;

	assert(t);
	assert(t->op == TypeStruct || t->op == TypeSyntax || t->op == TypeUnion);

	tl = hashget(t->tagshash, name);
	for(; tl; tl=tl->next){
		tag = tl->x;
		if(isvisiblescope(tag->declscope, scope))
			return tag;
	}

	// Lazy syntax type resolution.
	if(t->op == TypeSyntax || t->op == TypeUnion){
		tag = typesubname(t, name);
		if(tag->op != TypeSubname)
			return tag;
	}

	return nil;
}

// Lookup an attribute in a struct.
Var*
lookupattr(Type *t, Name name, Scope *scope, int dosuper)
{
	Var *v;

	t = unfoldstruct(t);
	// Self
	for(List<Var*> *l=t->vars; l; l=l->next){
		v = l->x;
		if(v->name == name){
			if((v->attrfn || v->isbuiltinattr) && isvisiblescope(v->declscope, scope))
				return v;
		}
	}
	// Super
	if(dosuper && t->super)
		return lookupattr(t->super, name, scope, 1);
	return nil;
}

// Find the base struct (highest supertype) of struct 't'.
Type*
rootstruct(Type *t)
{
	assert(t && (t->op == TypeStruct || t->op == TypeSyntax || t->op == TypeUnion));
	assert(!t->extends);
	while(t->super)
		t = t->super;
	return t;
}

// Find the owner struct of some tag struct 't', or just
// 't' if it has no supertype.  This is not necessarily
// the same as rootstruct() (see note above).
Type*
lookuptagowner(Type *type)
{
	Type *t;

	t = type;
	assert(t);
	assert(t->op == TypeStruct || t->op == TypeSyntax || t->op == TypeUnion);
	assert(!t->extends);
	for(; t; t=t->super)
		if(lookuptag(t, type->name, type->declscope) == type)
			return t;
	return nil;
}

// ========================================================
// Grammar types.

// Each of these functions takes a syntax type t and returns
// the requested super type (a Grammar, Sym, or Rule).

Type*
gramsupertype(Type *type)
{
	// Go up to the type with no super.
	if(!type)
		return nil;
	while(type->super)
		type = type->super;
	return type;
}

Type*
symsupertype(Type *type)
{
	// Go up to the type with one super.
	if(!type || !type->super)
		return nil;
	while(type->super->super)
		type = type->super;
	return type;
}

Type*
rulesupertype(Type *type)
{
	// Go up to the type with two supers.
	if(!type || !type->super || !type->super->super)
		return nil;
	while(type->super->super->super)
		type = type->super;
	return type;
}

// Replace the grammar type of 'type' with 'tgram'.
// Returns nil if impossible.
Type*
replacegramtype(Type *type, Type *tgram)
{
	if(!type)
		return nil;
	if(!type->super)
		return tgram;
	Type *s = replacegramtype(type->super, tgram);
	if(s == type->super)
		return type;
	return lookuptag(s, type->name, nil);
}

// Does this symname represent an Ast list?
int
isastlistsymname(Name name, Name *inner)
{
	if(name == nil)
		return 0;
	if(!strchr(name,'*') && !strchr(name,'+'))
		return 0;
	if(inner){
		char buf[512];
		char *ch;
		// name* --> name
		// name+ --> name
		// name[sep]* --> name
		// name[sep]+ --> name
		strncpy(buf, name, sizeof(buf) - 1);
		if((ch = strchr(buf,'[')) != nil)
			*ch = '\0';
		if((ch = strchr(buf,'*')) != nil)
			*ch = '\0';
		if((ch = strchr(buf,'+')) != nil)
			*ch = '\0';
		*inner = nameof(buf);
	}
	return 1;
}

// Does this symname represent an Ast option?
int
isastoptionsymname(Name name, Name *inner)
{
	int len;

	if(name == nil || (len = strlen(name)) == 0)
		return 0;
	if(name[len-1] != '?')
		return 0;
	if(inner){
		char buf[512];
		assert(len < (int)sizeof buf);
		// name? --> name
		strncpy(buf, name, sizeof(buf) - 1);
		buf[len-1] = '\0';
		*inner = nameof(buf);
	}
	return 1;
}

// Is this a type ptr Ast.sym, where 'sym' is a list symbol?
int
isastlisttype(Type *type, Type **sub)
{
	Name elemname;

	if(!isptrast(type))
		return 0;
	if((type = symsupertype(type->left)) == nil)
		return 0;
	// Grok the symname
	if(isastlistsymname(type->name, &elemname)){
		// Construct the list element type, if requested
		if(sub){
			type = lookuptag(type->super, elemname, globalscope);
			assert(type);
			*sub = typelist(typeptr(type));
		}
		return 1;
	}
	return 0;
}

// Is this a type ptr Ast.sym, where 'sym' is an option?
int
isastoptiontype(Type *type, Type **sub)
{
	Name elemname;

	if(!isptrast(type))
		return 0;
	if((type = symsupertype(type->left)) == nil)
		return 0;
	// Grok the symname
	if(isastoptionsymname(type->name, &elemname)){
		// Construct the option type, if requested
		if(sub){
			type = lookuptag(type->super, elemname, globalscope);
			assert(type);
			*sub = typeptr(type);
		}
		return 1;
	}
	return 0;
}


